Package: newmark
Type: Package
Title: Uncertainty Analysis in Dynamic Site and Slope Response
Version: 1.1.0
Authors@R: 
    person(given = "Alejandro",
           family = "Verri Kozlowski",
           email = "averri@fi.uba.ar",
           role = c("aut", "cre", "cph"))
Description: Implements a four-stage pipeline for probabilistic seismic
  performance analysis of slopes and embankments. The package takes a
  uniform-hazard spectrum at multiple return periods as input (any
  source) and produces: (1) synthetic soil profile generation and
  fundamental period estimation from USCS classification via Ishihara's
  small-strain shear-modulus model and the inhomogeneous truncated
  shear-beam theory of Gazetas and Dakoulas; (2) nonlinear site
  amplification using the Seyhan & Stewart (2014) model
  <doi:10.1193/063013EQS181M>, with inter-period correlation via
  Baker & Jayaram (2008) <doi:10.1193/1.2857544>; (3) Monte Carlo
  ensemble of six empirical Newmark sliding-block displacement models
  (Ambraseys & Menu (1988) <doi:10.1002/eqe.4290160704>,
   Jibson (2007) <doi:10.1016/j.enggeo.2007.01.013>,
   Saygili & Rathje (2008) <doi:10.1061/(ASCE)1090-0241(2008)134:6(790)>,
   Bray & Travasarou (2007) <doi:10.1061/(ASCE)1090-0241(2007)133:4(381)>,
   Bray & Macedo (2017) <doi:10.1016/j.soildyn.2017.05.024>,
   and the Bray and Macedo shallow-crustal update) with coherent
  correlated draws; (4) log-log inversion to the performance-based
  seismic coefficient kmax at user-specified displacement targets. All
  outputs are 'data.table' objects.
License: MIT + file LICENSE
Encoding: UTF-8
Language: en-US
LazyData: true
LazyDataCompression: xz
URL: https://github.com/averriK/newmark,
        https://averriK.github.io/newmark/
BugReports: https://github.com/averriK/newmark/issues
RoxygenNote: 7.3.2
Depends: R (>= 4.2)
Imports: data.table, digest, Hmisc, mvtnorm, randomForest, rpart,
        readxl, stats, stringr, triangle, utils
Suggests: testthat (>= 3.0.0), knitr, rmarkdown, devtools
VignetteBuilder: knitr
Config/testthat/edition: 3
NeedsCompilation: no
Config/roxygen2/version: 8.0.0
Packaged: 2026-05-14 01:46:40 UTC; averrik
Author: Alejandro Verri Kozlowski [aut, cre, cph]
Maintainer: Alejandro Verri Kozlowski <averri@fi.uba.ar>
Repository: CRAN
Date/Publication: 2026-05-19 08:20:02 UTC
Built: R 4.6.0; ; 2026-05-19 11:20:15 UTC; unix

# Generated by roxygen2: do not edit by hand

export(Dn_AM88)
export(Dn_BM17)
export(Dn_BM19)
export(Dn_BT07)
export(Dn_JB07)
export(Dn_SR08)
export(F_ST17)
export(SIDtoVs30)
export(SaF_ST17)
export(Vs30toSID)
export(approx.spline)
export(buildGMDP)
export(buildQSpline)
export(checkUHS)
export(designUHS)
export(fitDnCurve)
export(fitDnModel)
export(fitModel.Ts)
export(fitSaF)
export(geSiteTable)
export(getCylinderRoots)
export(getDnKy)
export(getKyLimits)
export(getSiteProperties)
export(interpolateSaTable)
export(invertDnDraws)
export(rhoBJ)
export(sampleSaCorr)
import(data.table)
import(digest)
import(randomForest)
import(rpart)
import(stats)
import(stringr)
import(utils)
importFrom(mvtnorm,rmvnorm)
importFrom(stats,pnorm)
importFrom(stats,predict)
importFrom(stats,splinefun)
importFrom(stats,weighted.mean)
importFrom(triangle,rtriangle)

# newmark

<!-- badges: start -->
[![R Version](https://img.shields.io/badge/R-%3E%3D%204.2.0-blue)](https://www.r-project.org/)
[![CRAN status](https://www.r-pkg.org/badges/version/newmark)](https://CRAN.R-project.org/package=newmark)
[![CRAN downloads](https://cranlogs.r-pkg.org/badges/newmark)](https://CRAN.R-project.org/package=newmark)
<!-- badges: end -->

## Overview

The package implements probabilistic propagation of seismic hazard
uncertainty through the Newmark sliding-block displacement framework.
Two stages of variability — the joint distribution of rock-level
spectral demand from PSHA quantiles, and the conditional dispersion
of empirical site-amplification and sliding-block models — are
sampled via Gaussian copulas with mean-preserving recentering, and
propagated to the mean, dispersion and design quantiles of the
residual displacement Dn(ky) and the performance-based seismic
coefficient kmax(d*).

Inputs are uniform-hazard spectra (UHS) at multiple return periods,
produced by any external PSHA engine.
Outputs are tabulated kmax(d*) and the normalised pseudostatic
coefficient Kh = kmax / PGA at user-specified displacement targets,
intended for performance-based seismic stability of slopes,
embankments, tailings storage facilities and waste rock dumps.

## Installation

```r
# Stable release
install.packages("newmark")

# Development version
devtools::install_github("averriK/newmark")
```

## Getting started

Four vignettes ship with the package:

```r
vignette("newmark-quickstart",     package = "newmark")    # runnable hello world
vignette("pipeline",               package = "newmark")    # function-level overview
vignette("dynamic-site-response",  package = "newmark")    # Ts, stiffness profile, site amplification
vignette("ensemble-formulation",   package = "newmark")    # mathematical derivation
```

Function-level reference: `?fitSaF`, `?getDnKy`, `?fitDnCurve`,
`?invertDnDraws`, `?getSiteProperties`, `?geSiteTable`,
`?getCylinderRoots`, `?fitModel.Ts`, `?buildGMDP`, `?Vs30toSID`,
`?SIDtoVs30`, or `help(package = "newmark")`.

Rendered documentation: <https://averriK.github.io/newmark/>.

## Getting help

Open an issue at <https://github.com/averriK/newmark/issues>.

## Citation

If you use this package in a publication, please cite the companion paper:

```bibtex
@unpublished{VerriKozlowski2026,
  author = {Verri Kozlowski, Alejandro},
  title  = {Probabilistic estimation of {Newmark} displacements and seismic
            coefficients under hazard uncertainty},
  year   = {2026},
  note   = {Working paper}
}
```

To cite the software package itself:

```bibtex
@Manual{Verri2026,
  title  = {{newmark}: Uncertainty Analysis in Dynamic Site and Slope Response},
  author = {Verri Kozlowski, Alejandro},
  year   = {2026},
  note   = {R package version 1.1.0},
  url    = {https://github.com/averriK/newmark}
}
```

## License

MIT — see [LICENSE](LICENSE)

## Author

**Alejandro Verri Kozlowski** — averri@fi.uba.ar  
[ORCID 0000-0002-8535-1170](https://orcid.org/0000-0002-8535-1170)  
Universidad de Buenos Aires, Facultad de Ingeniería

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")

## ----load---------------------------------------------------------------------
library(newmark)
library(data.table)

uhs <- fread(system.file("extdata", "uhs.csv", package = "newmark"))
uhs[p %in% c("0.16", "mean", "0.84") & Tn <= 0.2, .(Tn, p, Sa)]

## ----params-------------------------------------------------------------------
# Ts: fundamental period of the sliding mass (s).
# In production, derived from getSiteProperties() + getCylinderRoots()
# using the soil USCS profile and slope geometry (Ishihara 1996,
# Gazetas & Dakoulas 1985). Here set to a representative value.
Ts <- 0.60

# Mw: moment magnitude from PSHA disaggregation.
Mw <- 6.8

# Ensemble weights (0 = model inactive).
weights <- c(AM88 = 1, JB07 = 0, BT07 = 1, SR08 = 1, BM17 = 0, BM19 = 1)

# Displacement targets (cm).
Da <- c(0.5, 2.5, 5.0, 25.0)

## ----fitDnCurve---------------------------------------------------------------
ky     <- getDnKy(uhs, Ts = Ts)
result <- fitDnCurve(
  uhs     = uhs,
  ky      = ky,
  Ts      = Ts,
  Mw      = Mw,
  NS      = 200,
  weights = weights
)

result$curve[IDn == "ensemble" & p == "mean", .(ky, Dn)]

## ----invertDnDraws------------------------------------------------------------
kmax <- invertDnDraws(result$draws, Da = Da, weights = weights)
kmax[p %in% c("0.16", "mean", "0.84")]

## ── newmark: CRAN release workflow ───────────────────────────────────────────
## Run each section manually and in order. Do NOT source() end-to-end.
## Every step must pass before advancing to the next.
##
## This script reflects the flow that produced the 1.1.0 submission on
## 2026-05-09. Update any version-specific defaults (Step 1, Step 5)
## as you go, but the structure should hold for future patch / minor
## releases.

## ── Step 0: pre-flight ────────────────────────────────────────────────────────
## Confirm you are on main and the working tree is clean.
system("git checkout main && git pull --ff-only")
system("git status")
system("git log --oneline -5")

## ── Step 1: bump version ──────────────────────────────────────────────────────
## Edit DESCRIPTION manually before running anything else.
## Versioning rules:
##   bug fix only           → patch:  1.0.x
##   new exported function  → minor:  1.x.0
##   removed/renamed param  → major:  x.0.0
## CRAN policy: no more than one submission every ~2 months.

## ── Step 2: update NEWS.md ────────────────────────────────────────────────────
## Prepend a new top-level entry above the previous version:
##   # newmark 1.x.y
##   ## Bug fixes
##   * <change 1>
##   ## Documentation
##   * <change 2>
## Keep entries terse — one line per change.

## ── Step 3: regenerate documentation ─────────────────────────────────────────
devtools::document()
## Verify: man/ files updated, NAMESPACE correct.

## ── Step 4: local CRAN check ──────────────────────────────────────────────────
devtools::check(document = FALSE, cran = TRUE, force_suggests = TRUE)
## Must exit with 0 ERRORs, 0 WARNINGs, 0 NOTEs (or justify each NOTE).
## Common NOTE to justify on a first submission: "New submission" or
## "Maintainer field".

## ── Step 5: edit cran-comments.md ─────────────────────────────────────────────
## The file already exists at the repo root and is in .Rbuildignore
## (does not ship inside the tarball). Open it and update:
##
##   ## Test environments
##   * local: <macOS / R version of the day>
##   * GitHub Actions: ubuntu-latest, macos-latest, windows-latest (R-release)
##   * R-hub v2: linux, windows, macos-arm64        # if Step 6 was run
##
##   ## R CMD check results
##   0 errors | 0 warnings | 0 notes               # confirm from Step 4
##
##   ## Reverse dependencies
##   None on CRAN.                                 # or list affected
##
##   ## Notes for reviewer
##   <ONLY on resubmissions or unusual situations — explain what
##   changed in response to prior reviewer feedback. Delete on the
##   next clean submission.>

## ── Step 6: cross-platform check (optional) ──────────────────────────────────
## (a) R-hub v2 — triggers the .github/workflows/rhub.yaml workflow.
##     Requires RHUB_TOKEN secret in the repo.
# rhub::rhub_check(platforms = c("linux", "windows", "macos-arm64"))
##
## (b) win-builder — accepts a tarball at win-builder.r-project.org.
# devtools::check_win_devel()
# devtools::check_win_release()
##
## Monitor R-hub at: https://github.com/averriK/newmark/actions

## ── Step 7: commit, tag, push ─────────────────────────────────────────────────
## Stage: DESCRIPTION + NEWS.md + cran-comments.md + any man/, NAMESPACE
## changes from Step 3. Commit as a release commit, tag, push main + tag,
## then ff-sync dev.
# system('git add DESCRIPTION NEWS.md cran-comments.md man/ NAMESPACE')
# system('git commit -m "chore: release vX.Y.Z"')
# system('git tag -a vX.Y.Z -m "newmark X.Y.Z"')
# system('git push origin main vX.Y.Z')
# system('git checkout dev && git merge --ff-only main && git push origin dev && git checkout main')

## ── Step 8: submit to CRAN ────────────────────────────────────────────────────
## MUST be run from an INTERACTIVE R session (RStudio, or `R` in a
## terminal — not `R -e` / `Rscript`). devtools::submit_cran() prompts
## once to confirm the maintainer email; in non-interactive contexts
## it aborts with "Called from non-interactive context".
##
## What it does:
## 1. Builds the tarball.
## 2. Uploads it to https://xmpalantir.wu.ac.at/cransubmit/.
## 3. CRAN-SUBMISSION at the repo root is updated with the new
##    Date / SHA stamp (commit this small change after the run).
## 4. Sends a confirmation email to the maintainer.
devtools::submit_cran()

## ── Step 9: confirm submission via email ─────────────────────────────────────
## Open the email from "CRAN <CRAN@R-project.org>" with subject
## "[CRAN-pretest] newmark X.Y.Z" or similar, and follow the link to
##   https://xmpalantir.wu.ac.at/cransubmit/conf_mail.php?code=...
##
## The page is "Step 3 of 3" — a confirmation form, NOT a re-upload.
## Tick the two checkboxes:
##   [ ] I have read the CRAN policies and would like to submit this
##       package to CRAN.
##   [ ] I have checked the submission using R CMD check --as-cran and
##       a current version of r-devel, as mandated by the CRAN
##       Repository Policy.
## Then click submit. The package is now in CRAN's review queue.

## ── Step 10: wait for CRAN response ──────────────────────────────────────────
## CRAN typically responds within 1-5 business days by email:
##   * accepted   → no further action; the package appears at
##                  https://cran.r-project.org/package=newmark
##   * pre-test   → automated checks failed; address and resubmit
##                  (incrementing patch is usually NOT required if you
##                  are still pre-acceptance).
##   * manual     → reviewer requested changes; address, edit
##                  cran-comments.md "Notes for reviewer" with what you
##                  changed, and resubmit.

## ── Step 11: post-acceptance ──────────────────────────────────────────────────
## 1. Verify badge resolves: https://www.r-pkg.org/badges/version/newmark
## 2. The pkgdown workflow auto-deploys https://averriK.github.io/newmark/
##    on every push to main; nothing to do unless docs need a rebuild.
## 3. Commit the CRAN-SUBMISSION update from Step 8.3 if not yet committed.

## ── newmark: developer workflow ──────────────────────────────────────────────
## Run sections manually as needed.  Do NOT source() this file end-to-end.
## For CRAN release workflow see inst/cran-release.R.

## ── Install ───────────────────────────────────────────────────────────────────
## From CRAN (stable release — update only when a new version is accepted):
# install.packages("newmark")

## From GitHub (latest commit on main — picks up every push immediately):
# remotes::install_github("averriK/newmark")
# pak::pak("averriK/newmark")   # alternative if pak is available

## ── Development cycle ─────────────────────────────────────────────────────────
## 1. Regenerate documentation (roxygen2 → NAMESPACE + man/)
devtools::document()

## 2. Local CRAN check
devtools::check(document = FALSE, cran = TRUE, force_suggests = TRUE)

## 3. Local install with vignettes
devtools::install(build_vignettes = TRUE, dependencies = TRUE,
                  build = TRUE, upgrade = TRUE)

## 4. Render pkgdown site locally (optional sanity check before pushing)
# pkgdown::build_site()

# nolint start
#  Displacement models (empirical equations) --------------------------------
#' Ambraseys & Menu (1988) rigid sliding-block model
#' @param PGA Peak ground acceleration (g).
#' @param ky  Yield acceleration (g).
#' @return data.table(muLnD, sdLnD, ID = "AM88")
#' @examples
#' Dn_AM88(PGA = 0.5, ky = 0.1)
#' @export
Dn_AM88 <- function(PGA, ky) {
    r <- ky / PGA
    r <- pmin(r, 0.9999) # keep (1 - r) > 0
    muLog10D <- 0.90 + log10((1 - r)^2.53 * r^-1.09)
    sdLog10D <- 0.30

    data.table(
        muLnD = muLog10D * log(10),
        sdLnD = sdLog10D * log(10),
        ID = "AM88"
    )
}


#' Jibson (2007) empirical model
#' @inheritParams Dn_AM88
#' @param AI Arias Intensity (m/s); if NULL, back-calculated from PGA.
#' @return data.table(muLnD, sdLnD, ID = "JB07")
#' @examples
#' Dn_JB07(PGA = 0.5, ky = 0.1)
#' @export
Dn_JB07 <- function(PGA, ky, AI = NULL) {
    if (is.null(AI)) AI <- (PGA^1.9228) * exp(2.6109)
    r <- ky / PGA
    muLog10D <- 0.561 * log10(AI) - 3.833 * log10(r) - 1.474
    sdLog10D <- 0.616

    data.table(
        muLnD = muLog10D * log(10),
        sdLnD = sdLog10D * log(10),
        ID = "JB07"
    )
}

#' Saygili & Rathje (2008) model
#' @inheritParams Dn_JB07
#' @return data.table(muLnD, sdLnD, ID = "SR08")
#' @examples
#' Dn_SR08(PGA = 0.5, ky = 0.1)
#' @export
Dn_SR08 <- function(PGA, ky, AI = NULL) {
    if (is.null(AI)) AI <- (PGA^1.9228) * exp(2.6109)
    r <- ky / PGA
    muLnD <- 2.39 - 5.24 * r - 18.78 * r^2 + 42.01 * r^3 - 29.15 * r^4 -
        1.56 * log(PGA) + 1.38 * log(AI)
    sdLnD <- 0.46 + 0.56 * r

    data.table(
        muLnD = muLnD,
        sdLnD = sdLnD,
        ID = "SR08"
    )
}

#' Bray & Travasarou (2007) flexible sliding-block model
#' @param ky Yield acceleration (g).
#' @param Sa Spectral acceleration at 1.5 * Ts (g).
#' @param Ts Fundamental period (s).
#' @param Mw Moment magnitude.
#' @return data.table(muLnD, sdLnD, ID = "BT07")
#' @examples
#' Dn_BT07(ky = 0.1, Sa = 0.8, Ts = 0.3, Mw = 7.0)
#' @export
Dn_BT07 <- function(ky, Sa, Ts, Mw = 6.5) {
    lnSa <- log(Sa)
    lnky <- log(ky)

    muLnD <- -1.10 - 2.83 * lnky - 0.333 * lnky^2 +
        0.566 * lnky * lnSa + 3.04 * lnSa -
        0.244 * lnSa^2 + 0.278 * (Mw - 7)

    sdLnD <- 0.66

    data.table(
        muLnD = muLnD,
        sdLnD = sdLnD,
        ID = "BT07"
    )
}

#' Bray & Macedo (2017) - subduction/interface events
#' @param ky Yield acceleration (g).
#' @param Sa Spectral acceleration at 1.5 * Ts (g).
#' @param Ts Fundamental period (s).
#' @param Mw Moment magnitude (default 7.5).
#' @return data.table(muLnD, sdLnD, ID = "BM17")
#' @examples
#' Dn_BM17(ky = 0.1, Sa = 0.8, Ts = 0.3, Mw = 8.0)
#' @export
Dn_BM17 <- function(ky, Sa, Ts, Mw = 7.5) {
    lnSa <- log(Sa)

    if (Ts < 0.1) {
        c1 <- -5.864
        c2 <- -9.421
        c3 <- 0.0
    } else {
        c1 <- -6.896
        c2 <- 3.081
        c3 <- -0.803
    }

    a0 <- c1 + 0.550 * Mw + c2 * Ts + c3 * Ts^2 -
        3.353 * log(ky) - 0.390 * log(ky)^2
    a1 <- 3.060 + 0.538 * log(ky)
    a2 <- -0.225

    muLnD <- a0 + a1 * lnSa + a2 * lnSa^2
    sdLnD <- 0.73

    data.table(
        muLnD = muLnD,
        sdLnD = sdLnD,
        ID = "BM17"
    )
}

#' Bray & Macedo (2019, corrected 2023) - shallow-crustal events
#'
#' Multi-equation structure per BM23 (Soil Dynamics and Earthquake Engineering
#' 168, 107835, DOI 10.1016/j.soildyn.2023.107835):
#'   Eq (10): ordinary motions, PGV <= 115 cm/s
#'   Eq (6):  D100 max-component, PGV > 115 cm/s
#'   Eq (7):  D50  median-component, PGV > 115 cm/s
#' with a sub-regime split at PGV = 150 cm/s within Eqs (6) and (7).
#'
#' @param ky   Yield acceleration (g).
#' @param Ts   Fundamental period (s).
#' @param Sa   Spectral acceleration at 1.3 * Ts (g).
#' @param PGA  Peak ground acceleration (g).
#' @param PGV  Peak ground velocity (cm/s).
#' @param Mw   Moment magnitude (default 6.5).
#' @param NFC  Near-fault component: "D100" (default; max-component, slope
#'   within 45 deg of fault-normal) or "D50" (median-component, other
#'   orientations). Used only when PGV > 115 cm/s.
#' @return data.table(muLnD, sdLnD, ID = "BM19")
#' @examples
#' Dn_BM19(ky = 0.1, Ts = 0.3, Sa = 0.8, PGA = 0.5, PGV = 80, Mw = 7.0)
#' @export
Dn_BM19 <- function(ky, Ts, Sa, PGA, PGV, Mw = 6.5, NFC = "D100") {
    n <- length(Sa)
    if (!(length(PGA) == n && length(PGV) == n)) {
        stop("Sa, PGA, and PGV must have the same length")
    }
    if (!NFC %in% c("D100", "D50")) {
        stop("NFC must be 'D100' or 'D50'")
    }

    lnSa  <- log(Sa)
    lnky  <- log(ky)
    lnPGV <- log(PGV)

    Ordinary  <- PGV <= 115
    PulseLow  <- PGV >  115 & PGV <= 150
    PulseHigh <- PGV >  150

    # Eq (10): ordinary motions — identical to the pre-fix implementation
    A0Ord <- if (Ts < 0.1) {
        -4.684 + 0.603 * Mw - 9.471 * Ts - 2.482 * lnky - 0.244 * lnky^2
    } else {
        -5.981 + 0.603 * Mw + 3.223 * Ts - 0.945 * Ts^2 - 2.482 * lnky - 0.244 * lnky^2
    }
    MuOrd <- A0Ord + (2.649 + 0.344 * lnky) * lnSa - 0.090 * lnSa^2

    # Pulse equation coefficients (Eq 6 D100 or Eq 7 D50)
    if (NFC == "D100") {
        # Eq (6) — BM23 coefficients (verified from paper text)
        if (Ts < 0.1) {
            C1Low <- -6.724; C1High <-  1.991; c2 <- -2.744; c3 <-  0.000
        } else {
            C1Low <- -6.951; C1High <-  1.764; c2 <-  1.069; c3 <- -0.498
        }
        C4Low <- 1.547; C4High <- -0.097
        b1 <- 2.632; b2 <- 0.278; b3 <- 0.527; b4 <- 1.978; b5 <- 0.233
        Bm <- 0.01; SdPulse <- 0.56
    } else {
        # Eq (7) — BM23 coefficients (verified from paper text)
        if (Ts < 0.1) {
            C1Low <- -7.497; C1High <-  2.480; c2 <- -2.731; c3 <-  0.000
        } else {
            C1Low <- -7.718; C1High <- -0.369; c2 <-  1.031; c3 <- -0.480
        }
        C4Low <- 1.458; C4High <- 0.025
        b1 <- 2.931; b2 <- 0.319; b3 <- 0.584; b4 <- 2.261; b5 <- 0.241
        Bm <- 0.05; SdPulse <- 0.54
    }

    C1Vec <- ifelse(PulseLow, C1Low, ifelse(PulseHigh, C1High, NA_real_))
    C4Vec <- ifelse(PulseLow, C4Low, ifelse(PulseHigh, C4High, NA_real_))

    MuPulse <- C1Vec - b1 * lnky - b2 * lnky^2 + b3 * lnky * lnSa +
               b4 * lnSa - b5 * lnSa^2 + c2 * Ts + c3 * Ts^2 +
               Bm * Mw + C4Vec * lnPGV

    muLnD <- ifelse(Ordinary, MuOrd, MuPulse)
    sdLnD <- ifelse(Ordinary, 0.72, SdPulse)

    data.table(
        muLnD = muLnD,
        sdLnD = sdLnD,
        ID = "BM19"
    )
}

#' Cylinder Roots Data
#'
#' This dataset contains cylinder roots.
#'
#' @format A data frame with 307124 rows and 4 variables:
#' \describe{
#'   \item{n}{Integer: Root number (1-10)}
#'   \item{m}{Double: Inhomogeneity ratio (0-0.95)}
#'   \item{l}{Double: Aspect Ratio (0-0.4950)}
#'   \item{an}{Double: Cylinder roots}
#' }
#' @usage data(CylinderRoots)
#' @docType data
#' @keywords datasets
#' @name CylinderRoots
"CylinderRoots"

#' @title Convert Vs30 (m/s) to Site Class Identifier
#'
#' @description
#' Given one or more Vs30 values (in m/s), returns a character vector of site
#' class designations (\code{"A", "B", "BC", "C", "CD", "D", "DE", "E"}). The
#' classification thresholds are:
#' \itemize{
#'   \item A: Vs30 >= 1500
#'   \item B: 900 <= Vs30 < 1500
#'   \item BC: 640 <= Vs30 < 900
#'   \item C: 440 <= Vs30 < 640
#'   \item CD: 300 <= Vs30 < 440
#'   \item D: 210 <= Vs30 < 300
#'   \item DE: 150 <= Vs30 < 210
#'   \item E: 0 <= Vs30 < 150
#' }
#'
#' @param Vs30 Numeric vector. One or more Vs30 values in m/s.
#'
#' @return A character vector of the same length as \code{Vs30},
#'         with site class designations.
#' @export
#'
#' @examples
#' Vs30toSID(1500)  # returns "A"
#' Vs30toSID(c(120, 790, 3000, 455))
Vs30toSID <- function(Vs30) {
  if (!is.numeric(Vs30)) {
    stop("`Vs30` must be numeric.")
  }
  sapply(Vs30, .Vs30toSID_char)
}

# Internal helper that does the classification for a single numeric Vs30
.Vs30toSID_char <- function(v) { # nolint: object_name_linter.
  if (v >= 1500) {
    "A"
  } else if (v >= 900) {
    "B"
  } else if (v >= 640) {
    "BC"
  } else if (v >= 440) {
    "C"
  } else if (v >= 300) {
    "CD"
  } else if (v >= 210) {
    "D"
  } else if (v >= 150) {
    "DE"
  } else if (v >= 0) {
    "E"
  } else {
    stop("Vs30 < 0 is invalid. Value = ", v)
  }
}

#' Site's Fundamental Period
#'
#' @param VSm Vector. Shear Wave Velocity profile in m/s (vector)
#' @param hs Vector. Layer Thickness profile in m (vector)
#' @param zm Vector. Layer Coordinates profile in m (vector)
#'
#' @return Double. Site's fundamental period Ts in seconds
#' @export fitModel.Ts
#'
#' @import data.table
#'
fitModel.Ts <- function(VSm, hs, zm) { # nolint: object_name_linter.
  NL <- length(VSm)
  A <- vector(mode = "double", length = NL - 1)
  B <- vector(mode = "double", length = NL - 1)
  f <- vector(mode = "double", length = NL)
  Hs <- sum(hs)
  for (j in seq(1, NL)) {
    f[j + 1] <- f[j] + hs[j] * (Hs - zm[j]) / VSm[j]
    A[j] <- (VSm[j] * (f[j + 1] - f[j]))^2 / hs[j]
    B[j] <- hs[j] * (f[j + 1] + f[j])^2
  }
  ws2 <- 4 * sum(A) / sum(B)
  Ts <- (2 * pi / sqrt(ws2)) |> round(3)
  Ts
}

###############################################################################
# getDnKy.R — Scenario-specific ky grid for Dn(ky) evaluation
###############################################################################

#' Build a log-spaced ky grid for one scenario
#'
#' Derives the effective intensities PGA, Sa(1.3Ts), Sa(1.5Ts) from the
#' scenario UHS and constructs a log-spaced ky grid covering the calibrated
#' range of the displacement models.
#'
#' Default bounds come from the calibration ranges of the models in the
#' ensemble:
#' - lower: 0.01g (BM17 lowest calibration point)
#' - upper: max(PGA, 0.80g). The physical limit is ky < PGA (no slip below
#'   the yield acceleration); the 0.80g floor extends the grid for
#'   low-PGA scenarios up to BM17's calibration ceiling.
#'
#' @param uhs  data.table with columns Tn, p, Sa (one scenario)
#' @param Ts   numeric scalar, fundamental period (s)
#' @param kyN  integer, number of grid points (default 30)
#' @param pRef character, probability level for intensity anchor (default "mean")
#'
#' @return numeric vector of ky values (log-spaced)
#' @examples
#' uhs <- data.table::data.table(Tn = 0, p = "mean", Sa = 0.5)
#' getDnKy(uhs, Ts = 0.3)
#' @export
getDnKy <- function(uhs, Ts, kyN = 30L, pRef = "mean") {

  uhs <- data.table::as.data.table(uhs)[, .(Tn, p, Sa)]
  uhs[, p := as.character(p)]

  PGA <- uhs[Tn == min(Tn) & p == pRef, Sa]

  kyMin <- 0.01
  kyMax <- max(PGA, 0.80)

  exp(seq(log(kyMin), log(kyMax), length.out = kyN))
}

#' @importFrom stats weighted.mean
NULL

# Work around R CMD check "no visible binding" notes
utils::globalVariables(
  c("..ID", "AF", "w", "pga", "PGA", "N", "depth", "lat", "lon",
    "SaRock", "TnVal", "FValue", "SaF", "LnF", "muLnF", "sdLnF",
    "pnum", "lnSa", "Tn", "p", "Dn", "weight", "muLnD", "sdLnD",
    "LnD", "LnSaF", "muLnSaF", "sdLnSaF", "ID", "Sa", "SaF_SS17",
    "TR", "Vs30", "Vref", ".", "AEP", "IT", "POE", "iml", "imt",
    "ky", "s", "IDn", "kmax",
    # dsra globals
    "ValidSands", "ValidGravels", "ValidSilts", "ValidClays",
    "ValidOrganic", "ValidFines", "ValidCoarse", "ValidUSCS",
    "ValidGroups", "SiteTable", "VoidRatiosUSCS",
    "RelativeDensityRanges", "UnitWeightRanges", "ParticleSize",
    "USCS", "ShearModelParameters", "CylinderRoots", "SiteClass",
    "DensityID", "GroupID", "ModelID", "lo", "mo", "Hs", "Hw",
    "VSo", "Go", "an", "Fines", "Gravels", "Sands", "Water",
    "Clays", "Silts", "l", "m", "n",
    "Ts", "VS30", "Z500", "Z1000"
  )
)

###############################################################################
# fitDnCurve.R — Full Dn(ky) curve for one scenario
###############################################################################

#' Newmark displacement curve Dn(ky) for one scenario
#'
#' Evaluates ensemble and per-submodel Dn over a vector of ky values.
#' The scenario is sampled once and the same epsilon is reused across all ky
#' so that each realisation s represents a coherent physical curve Dn_s(ky).
#'
#' @param uhs      data.table with columns Tn, p, Sa (one scenario)
#' @param ky       numeric vector, yield accelerations (g)
#' @param Ts       numeric scalar, fundamental period (s)
#' @param Mw       numeric scalar, moment magnitude
#' @param NS       integer, Monte Carlo samples for curve quantiles and
#'   per-realisation draws.
#' @param Rrup     numeric scalar, rupture distance (km)
#' @param weights  named numeric vector, ensemble weights by IDn
#' @param NFC      character, near-fault component selector for model BM19:
#'   \code{"D100"} (maximum component, default) or \code{"D50"}.
#'   Ignored by all other submodels.
#'
#' @return list(curve, draws). curve = data.table(ky, p, Dn, IDn, w).
#'   draws = data.table(ky, s, Dn, IDn) — always populated (possibly
#'   empty when no model produces output). Dn in cm.
#' @examples
#' \dontrun{
#' uhs <- data.table::fread(
#'   system.file("extdata", "uhs.csv", package = "newmark"))
#' out <- fitDnCurve(uhs, ky = getDnKy(uhs, Ts = 0.3), Ts = 0.3, Mw = 7.5,
#'                   NS = 100, weights = c(AM88=1, BT07=1, SR08=1, BM17=1))
#' }
#' @export
fitDnCurve <- function(uhs, ky, Ts, Mw = 6.5, NS = 30L,
                       Rrup = 100, weights, NFC = "D100") {

  w <- as.numeric(weights)
  w <- w / sum(w)

  # Sample scenario once — shared across all ky and all models
  .sample <- .sampleDnScenario(uhs, Ts, NS, Rrup)

  # Generate coherent draws for every model across the full ky grid.
  # Each model gets its own pre-generated epsilon so realisations are
  # physical curves, not independent per-ky draws.
  Eps <- stats::rnorm(NS)
  AllDraws <- stats::setNames(
    lapply(names(weights), function(id) {
      .drawDnModelCurve(.sample, ky, Mw, Ts, id, NS, eps = Eps, NFC = NFC)
    }),
    names(weights)
  )

  ok <- names(weights)[vapply(
    AllDraws, function(d) !is.null(d) && nrow(d) > 0L, logical(1)
  )]

  if (!length(ok)) {
    EmptyCurve <- data.table::data.table(
      ky = numeric(), p = character(), Dn = numeric(),
      IDn = character(), w = numeric()
    )
    EmptyDraws <- data.table::data.table(
      ky = numeric(), s = integer(), Dn = numeric(), IDn = character()
    )
    return(list(curve = EmptyCurve, draws = EmptyDraws))
  }

  WOk <- w[match(ok, names(weights))]

  OUT <- vector("list", length(ky))

  for (j in seq_along(ky)) {
    k <- ky[j]

    LIST <- lapply(ok, function(id) {
      d <- AllDraws[[id]][ky == k, Dn]
      if (!length(d)) NULL else d
    })
    OkMask <- !vapply(LIST, is.null, logical(1))
    if (!any(OkMask)) next

    ListJ  <- LIST[OkMask]
    WOkJ   <- WOk[OkMask]
    IdsOkJ <- ok[OkMask]

    # Per-model curve: quantiles + mean
    AUX <- data.table::rbindlist(lapply(seq_along(IdsOkJ), function(i) {
      DT <- .summarizeDn(ListJ[[i]], .sample$pVals)
      DT[, `:=`(IDn = IdsOkJ[i], w = WOkJ[i])]
      DT
    }), use.names = TRUE)

    # Ensemble curve at this ky
    L <- .aggregateWeighted(ListJ, WOkJ, .sample$pVals)
    L[, `:=`(IDn = "ensemble", w = NA_real_)]

    DT <- data.table::rbindlist(list(L, AUX), use.names = TRUE)
    DT[, ky := k]
    OUT[[j]] <- DT
  }

  Curve <- data.table::rbindlist(Filter(Negate(is.null), OUT), use.names = TRUE)
  data.table::setcolorder(Curve, c("ky", "p", "Dn", "IDn", "w"))

  Draws <- data.table::rbindlist(AllDraws[ok], use.names = TRUE)

  list(curve = Curve, draws = Draws)
}

#' MCER / Design elastic spectrum for a Vs30/Vref slice (ASCE 7-22)
#'
#' The input `UHS` slice must contain at least `Tn` and `SaF`.
#' Typical usage passes a single TR and `p == "mean"`.
#'
#' @param UHS data.table slice from UHSTable (single TR, p == "mean").
#' @param TL  numeric, long-period transition (seconds). Default 8.
#' @param spectrum character, "MCER" (default) or "DESIGN" (2/3 * MCER).
#'
#' @return data.table with columns `Tn, SaF`.
#' @examples
#' uhs <- data.table::data.table(Tn = c(0, 0.2, 0.5, 1, 2), SaF = c(0.4, 1.0, 0.9, 0.6, 0.3))
#' designUHS(uhs)
#' @export
designUHS <- function(UHS, TL = 8, spectrum = c("MCER", "DESIGN")) {
  spectrum <- match.arg(spectrum)

  # Sort and extract raw vectors
  UHS <- UHS[order(Tn)]
  Tn0 <- as.numeric(UHS$Tn)
  Sa0 <- as.numeric(UHS$SaF)

  # Drop NA/Inf and collapse duplicates in Tn using the maximum SaF
  ok <- is.finite(Tn0) & is.finite(Sa0)
  Tn0 <- Tn0[ok]
  Sa0 <- Sa0[ok]
  if (length(Tn0) < 1L) {
    stop("designUHS(): no valid (Tn, SaF) after cleaning")
  }
  if (any(duplicated(Tn0))) {
    DTu <- data.table::data.table(Tn = Tn0, Sa = Sa0)[, .(Sa = max(Sa, na.rm = TRUE)), by = Tn]
    Tn0 <- DTu$Tn
    Sa0 <- DTu$Sa
  }
  if (length(Tn0) < 2L) {
    stop("designUHS(): need >= 2 (Tn, SaF) points after cleaning")
  }

  # Safe interpolator (linear in T; ASCE uses linear T-spacing)
  fSa <- stats::approxfun(Tn0, Sa0, rule = 2, ties = "mean")

  # Rock ordinates (already site-adjusted if SaF came from sdLnSaFC1)
  Ss <- fSa(0.2)  # S_s
  S1 <- fSa(1.0)  # S_1
  if (!is.finite(Ss) || !is.finite(S1)) {
    stop("designUHS(): invalid Ss or S1 from interpolation")
  }

  # Site factors (set to 1 here; apply Fa/Fv upstream if desired)
  Fa  <- 1
  Fv  <- 1
  SMS <- Fa * Ss
  SM1 <- Fv * S1

  # Corner periods and evaluation grid
  Ts <- SM1 / SMS
  if (!is.finite(Ts) || Ts <= 0) stop("designUHS(): Ts must be positive")
  if (!is.finite(TL) || TL <= 0) stop("designUHS(): TL must be positive")

  T0 <- 0.2 * Ts
  TnDecay1 <- exp(seq(log(max(Ts, 1e-6)), log(TL),          length.out = 20))
  TnDecay2 <- exp(seq(log(TL),               log(max(TL, 5)), length.out = 20))
  Tn <- sort(unique(c(0, T0, Ts, TL, TnDecay1, TnDecay2)))

  # MCER or DESIGN spectrum
  if (tolower(spectrum) == "mcer") {
    SaF <- piecewiseSa(t = Tn, sds = SMS, sd1 = SM1, TL = TL)
  } else {
    SaF <- piecewiseSa(t = Tn, sds = (2 / 3) * SMS, sd1 = (2 / 3) * SM1, TL = TL)
  }

  data.table::data.table(Tn = Tn, SaF = SaF)
}

# Piecewise ASCE 7-22 spectrum generator (SDS/SD1 form)
piecewiseSa <- function(t, sds, sd1, TL = 8) {
  Ts <- sd1 / sds
  T0 <- 0.2 * Ts
  t  <- pmax(t, 1e-6)
  ifelse(
    t < T0,  sds * (0.4 + 0.6 * t / T0),
    ifelse(
      t <= Ts, sds,
      ifelse(
        t <= TL, sd1 / t,
        sd1 * TL / t^2
      )
    )
  )
}

#' Characteristic root of an inhomogeneous truncated shear beam
#'
#' Returns the eigenvalue \code{an} of mode \code{no} for a Gazetas &
#' Dakoulas (1985) inhomogeneous truncated shear beam parameterised by
#' inhomogeneity ratio \code{mo} and truncation ratio \code{lo}. The
#' value is interpolated from the precomputed `CylinderRoots` table
#' using the requested model (linear glm, non-linear glm, decision
#' tree, or random forest).
#'
#' @param mo Double. Inhomogeneity ratio of the shear-modulus profile.
#'   \code{mo = 0} is a homogeneous column. Range 0 to 0.95.
#' @param lo Double. Truncation ratio (top-radius / bottom-radius);
#'   \code{lo = 0} is a fully truncated cone. Range 0 to 0.495.
#' @param no Integer. Mode number, 1..8. Default 1 (fundamental).
#' @param model Character. Interpolation model: \code{"lm"} (linear
#'   GLM), \code{"nlm"} (non-linear GLM with interactions, default),
#'   \code{"dt"} (decision tree), or \code{"rf"} (random forest).
#' @param extrapolate Logical. If \code{TRUE} (default), allow
#'   \code{mo}/\code{lo} outside the table support and warn; if
#'   \code{FALSE}, clamp to the table maxima.
#' @param OSF Double. Outlier scale factor used by the random-forest
#'   model to shrink the training set around the query point. Default
#'   0.10.
#'
#' @return Double. Characteristic root \code{an} for the requested
#'   mode and parameters.
#' @export getCylinderRoots
#'
#' @import data.table
#' @import stats
#' @import rpart
#' @import randomForest
#'
#' @examples
#' getCylinderRoots(mo = 0.45, lo = 0.20, no = 1, model = "nlm")

getCylinderRoots <- function(mo,lo,no=1,model="nlm",extrapolate=TRUE,OSF=0.10){

  # Ensure the lazy-loaded dataset is bound in this scope. Without
  # this call, bare-name references below resolve via the package
  # namespace under `library(newmark)` but can fail to resolve when
  # the package is loaded via `pkgload::load_all()` (the path pkgdown
  # takes when install = FALSE), producing
  #   "object 'CylinderRoots' not found".
  utils::data("CylinderRoots", package = "newmark", envir = environment())

  OK <- exists("CylinderRoots") & !is.null(CylinderRoots)
  stopifnot(OK)
  . <- NULL
  # Check ranges
  if(!(no %in% seq(min(CylinderRoots$n), max(CylinderRoots$n)) )){
    stop(sprintf("no=%d .Wrong mode number",no))}
  DATA <- CylinderRoots[n==no,-c("n")] |> unique()
  # Check if you were lucky
  if(nrow(DATA[l==lo & m==mo])>0){
    VALUE <- DATA[l==lo & m==mo]$an
    return(VALUE)
  }
  # Check ranges
  LoMax <- max(DATA$l)
  if(lo>LoMax & extrapolate==FALSE){
    lo <- min(lo,LoMax)
  }
  
  MoMax <- max(DATA$m)
  if(mo>MoMax & extrapolate==FALSE){
    mo <- min(mo,MoMax)
  }
  
  # Linear interpolation
  if(model=="lm"){
    MODEL <- glm(an ~l+m+l2+m2,data=DATA[,.(an,l,m,l2=l^2,m2=m^2)])
    NEWDATA <- list(l=lo,m=mo,l2=lo^2,m2=mo^2)
    VALUE <- predict(MODEL,newdata=NEWDATA)
    
  }
  # Non-Linear interpolation. Mixed effects. Best 
  if(model=="nlm"){
    MODEL <- glm(an ~l+m+l2+m2+l*m+l2*m2,data=DATA[,.(an,l,m,l2=l^2,m2=m^2)])
    NEWDATA <- list(l=lo,m=mo,l2=lo^2,m2=mo^2)
    VALUE <- predict(MODEL,newdata=NEWDATA)
    
  }
  
  # Decision Trees
  if(model=="dt"){
    MODEL <- rpart(an ~ l+m,data=DATA)
    NEWDATA <- list(l=lo,m=mo)
    VALUE <- predict(MODEL,newdata=NEWDATA)
    }
  
  
  # Random Forest
  if(model=="rf"){
    # Reduce the dataset
    DATA <- DATA[ between(m,(1-OSF)*mo,(1+OSF)*mo) & between(l,(1-OSF)*lo,(1+OSF)*lo)]
    MODEL <- randomForest(an ~ l+m,data=DATA,importance=FALSE,proximity=FALSE)
    NEWDATA <- list(l=lo,m=mo)
    VALUE <- predict(MODEL,newdata=NEWDATA)
  }
  if(lo>LoMax & mo<=MoMax & extrapolate==TRUE){
    warning(sprintf("lo=%g is higher than LoMax=%g. The rooth an=%g is a predicted value that may not satisfy the boundary conditions of characteristics (cylinder) equation. Use this value with proper judgement.",lo,LoMax,VALUE))
  }
  if(mo>MoMax & lo<=LoMax & extrapolate==TRUE){
    warning(sprintf("mo=%g is higher than MoMax=%g. The root an=%g is a predicted value that may not satisfy the boundary conditions of characteristics (cylinder) equation. Use this root value with proper judgement.",mo,MoMax,VALUE))
  }
  if(mo>MoMax & lo>LoMax & extrapolate==TRUE){
    warning(sprintf("mo=%g and lo=%g  are higher than MoMax=%g and LoMax=%g, respectively. The root an=%g is a predicted value that may not satisfy the boundary conditions of characteristics (cylinder) equation. Use this root value with proper judgement.",mo,lo,MoMax,LoMax,VALUE))
  }
  return(VALUE)
}

#' Convert Site Class to Vs30 in m/s
#'
#' @param SID Site Class
#'
#' @return Vs30 in m/s
#' @export SIDtoVs30
#' @import data.table
#'
#' @examples
#'
#' SIDtoVs30("A")
#'
#' SIDtoVs30("AB")
#'
#' SIDtoVs30(c("BC","C","CD","D"))
#'
#'
SIDtoVs30 <- function(SID = NULL) {
  OK <- all(toupper(SID) %in% c("A", "AB", "B", "BC", "C", "CD", "D", "DE", "E"))
  stopifnot(OK)
  sapply(SID, SIDtoVs30.char)
}

SIDtoVs30.char <- function(SID) { # nolint: object_name_linter.
  switch(EXPR = toupper(SID),
    "A" = 1500,
    "AB" = 1500,
    "B" = 1200,
    "BC" = 760,
    "C" = 540,
    "CD" = 370,
    "D" = 255,
    "DE" = 180,
    "E" = 150
  )
}

###############################################################################
# getKyLimits.R — Calibration ky range per Newmark displacement model
###############################################################################

#' Get calibrated ky range for each displacement model
#'
#' Returns the ky bounds within which each model's regression is calibrated.
#' For rigid models (AM88, SR08) the bounds are relative to PGA.
#' For flexible models the bounds are absolute.
#'
#' @param PGA numeric scalar, peak ground acceleration (g). Required for
#'   rigid model limits.
#'
#' @return data.table with columns IDn, ky.lo, ky.hi
#' @examples
#' getKyLimits(PGA = 0.5)
#' @export
getKyLimits <- function(PGA) {
  data.table::data.table(
    IDn   = c("AM88",  "JB07", "SR08", "BT07", "BM17", "BM19"),
    ky.lo = c(0.001,   0.05,   0.05,   0.02,   0.01,   0.01),
    ky.hi = c(PGA,     0.40,   0.30,   0.40,   0.80,   0.80)
  )
}

###############################################################################
# fitDnModel.R — Single-submodel Newmark displacement + internal helpers
###############################################################################

#' Newmark displacement for a single submodel and scenario
#'
#' @param uhs   data.table with columns Tn, p, Sa (one scenario)
#' @param ky    numeric scalar, yield acceleration (g)
#' @param Ts    numeric scalar, fundamental period (s)
#' @param Mw    numeric scalar, moment magnitude
#' @param NS    integer, Monte Carlo samples
#' @param Rrup  numeric scalar, rupture distance (km)
#' @param IDn   character, submodel identifier
#' @param NFC   character, near-fault component selector for model BM19:
#'   \code{"D100"} (maximum component, default) or \code{"D50"}.
#'   Ignored by all other submodels.
#'
#' @return data.table with columns p, Dn, IDn. Dn in cm.
#' @examples
#' \dontrun{
#' uhs <- data.table::fread(
#'   system.file("extdata", "uhs.csv", package = "newmark"))
#' fitDnModel(uhs, ky = 0.1, Ts = 0.3, Mw = 7.5, NS = 100, IDn = "AM88")
#' }
#' @export
fitDnModel <- function(uhs, ky, Ts, Mw = 6.5, NS = 30L, Rrup = 100,
                       IDn, NFC = "D100") {

  .sample <- .sampleDnScenario(uhs, Ts, NS, Rrup)
  Dn      <- .drawDnModel(.sample, ky, Ts, Mw, IDn, NFC = NFC)
  if (is.null(Dn)) {
    return(data.table::data.table(p = character(), Dn = numeric(), IDn = character()))
  }
  DT <- .summarizeDn(Dn, .sample$pVals)
  DT[, IDn := IDn]
  DT
}


# --- Internal helpers (not exported, shared with fitDnCurve) ----------------

.sampleDnScenario <- function(uhs, Ts, NS, Rrup) {
  uhs <- data.table::as.data.table(uhs)[, .(Tn, p, Sa)]
  uhs[, p := as.character(p)]

  periods <- c(0, 1.3 * Ts, 1.5 * Ts)

  LIST <- lapply(periods, function(Ti) {
    if (Ti %in% uhs$Tn) uhs[Tn == Ti]
    else interpolateSaTable(uhs, Tn = Ti)
  })
  SaTable <- data.table::rbindlist(LIST, use.names = TRUE)

  J <- length(periods)
  C <- diag(J)
  for (j in seq_len(J)) for (k in seq_len(J)) {
    if (j != k) {
      Tj <- if (periods[j] == 0) 0.01 else periods[j]
      Tk <- if (periods[k] == 0) 0.01 else periods[k]
      C[j, k] <- rhoBJ(Tj, T0 = Tk)
    }
  }

  SaMatrix <- .sampleSaCorrFull(SaTable, Tn = periods, rho = C, NS = NS)

  PGA  <- SaMatrix[, 1L]
  Sa13 <- SaMatrix[, 2L]
  Sa15 <- SaMatrix[, 3L]
  PGV  <- (PGA^1.0529) * exp(0.1241) * 100
  AI   <- (PGA^1.9228) * exp(2.6109)

  x <- suppressWarnings(as.numeric(uhs[p != "mean", p]))
  pVals <- sort(unique(x[is.finite(x)]))

  list(PGA = PGA, Sa13 = Sa13, Sa15 = Sa15, PGV = PGV, AI = AI,
       pVals = pVals)
}


.drawDnModel <- function(.sample, ky, Ts, Mw, IDn, eps = NULL, NFC = "D100") {
  spec <- tryCatch({
    switch(IDn,
           AM88 = Dn_AM88(.sample$PGA, ky),
           JB07 = Dn_JB07(.sample$PGA, ky, AI = .sample$AI),
           SR08 = Dn_SR08(.sample$PGA, ky, AI = .sample$AI),
           BT07 = Dn_BT07(ky = ky, Sa = .sample$Sa15, Ts = Ts, Mw = Mw),
           BM17 = Dn_BM17(ky = ky, Sa = .sample$Sa15, Ts = Ts, Mw = Mw),
           BM19 = Dn_BM19(ky = ky, Ts = Ts, Sa = .sample$Sa13,
                          PGA = .sample$PGA, PGV = .sample$PGV, Mw = Mw,
                          NFC = NFC),
           stop("Unsupported IDn: ", IDn))
  }, error = function(e) NULL)

  if (is.null(spec)) return(NULL)
  if (is.null(eps)) eps <- stats::rnorm(length(.sample$PGA))
  exp(spec$muLnD + spec$sdLnD * eps)
}


# Draw coherent Dn(ky) curve for one model — same epsilon across all ky.
# If eps is supplied (shared across models), all models use the same
# standard-normal vector; each still scales by its own sigma_lnD.
.drawDnModelCurve <- function(.sample, ky, Mw, Ts, IDn, NS, eps = NULL,
                              NFC = "D100") {
  if (is.null(eps)) eps <- stats::rnorm(NS)
  LIST <- lapply(seq_along(ky), function(j) {
    Dn <- .drawDnModel(.sample, ky[j], Ts, Mw, IDn, eps = eps, NFC = NFC)
    if (is.null(Dn)) return(NULL)
    data.table::data.table(ky = ky[j], s = seq_len(NS), Dn = Dn, IDn = IDn)
  })
  data.table::rbindlist(Filter(Negate(is.null), LIST), use.names = TRUE)
}


# Summarize a draw table (ky, s, Dn) to quantiles + mean per ky.
.summarizeDnDraws <- function(draws, pVals) {
  draws[, .summarizeDn(Dn, pVals), by = .(ky)]
}


.summarizeDn <- function(Dn, pVals) {
  Q <- data.table::data.table()
  if (length(pVals)) {
    q <- stats::quantile(Dn, probs = pVals, na.rm = TRUE, names = FALSE)
    Q <- data.table::data.table(p = as.character(pVals), Dn = as.numeric(q))
  }
  M <- data.table::data.table(p = "mean", Dn = mean(Dn, na.rm = TRUE))
  data.table::rbindlist(list(Q, M), use.names = TRUE)
}


.aggregateWeighted <- function(LIST, w, pVals) {
  x  <- unlist(LIST)
  wt <- rep(w, each = length(LIST[[1L]]))

  Q <- data.table::data.table()
  if (length(pVals)) {
    q <- Hmisc::wtd.quantile(x, weights = wt,
                             probs = pVals, type = "quantile",
                             na.rm = TRUE)
    Q <- data.table::data.table(p = as.character(pVals), Dn = as.numeric(q))
  }
  M <- data.table::data.table(
    p  = "mean",
    Dn = Hmisc::wtd.mean(x, weights = wt, na.rm = TRUE)
  )
  data.table::rbindlist(list(Q, M), use.names = TRUE)
}

#' Check monotonicity and duplicates in UHSTable
#'
#' Ensures that, for every combination
#' \code{TR, Vs30, Vref, ID, Tn}, there is exactly
#' one row per probability \code{p} and that
#' \code{Sa(p)} is non‑decreasing.
#'
#' @param UHSTable data.table with columns
#'   \code{TR, Vs30, Vref, ID, Tn, p, Sa}.
#' @param epsMonot numeric tolerance (default 1e-6).
#'
#' @return invisibly \code{TRUE}; stops on error.
#' @examples
#' \dontrun{
#' # Requires a fully populated UHSTable produced by buildGMDP() or an
#' # equivalent upstream pipeline (must include the ID column that
#' # disaggregates per-quantile rows). Not bundled with the package.
#' checkUHS(UHSTable)
#' }
#' @export
checkUHS <- function(UHSTable, epsMonot = 1e-6) {
  stopifnot(inherits(UHSTable, "data.table"))

  dup <- UHSTable[p != "mean",
    .N,
    by = .(TR, Vs30, Vref, ID, Tn, p)
  ][N > 1L]
  if (nrow(dup)) {
    stop(
      sprintf(
        "checkUHS(): %d duplicados de cuantil.",
        nrow(dup)
      ),
      call. = FALSE
    )
  }

  bad <- UHSTable[p != "mean",
    {
      s <- Sa[order(p)]
      if (any(diff(s) < -epsMonot)) .SD
    },
    by = .(TR, Vs30, Vref, ID, Tn)
  ]

  if (nrow(bad)) {
    stop(
      sprintf(
        "checkUHS(): ERROR",
        uniqueN(bad, by = .(TR, Vs30, Vref, ID, Tn))
      ),
      call. = FALSE
    )
  }

  invisible(TRUE)
}

# Internal utilities from dsra

#' @keywords internal
#' @noRd
.mapLayers <- function(UID, hs, USCS = ValidUSCS) {
  DT <- data.table::data.table(UID, hs)
  Hs <- sum(hs)
  DT <- DT[, list(hs = round(sum(hs) / Hs * 100)), by = UID]
  DT <- data.table::dcast(DT, formula = "1 ~ UID", value.var = "hs", fill = 0)
  DT[, (USCS) := lapply(USCS, function(u) ifelse(u %in% names(DT), DT[[u]], 0))]
  DT[, .SD, .SDcols = USCS]
}

#' @keywords internal
#' @noRd
.find <- function(V, X) {
  V <- sort(unique(V))
  X <- sort(unique(X))
  results <- numeric(length(X))
  for (i in seq_along(X)) {
    x <- X[i]
    if (x %in% V) {
      results[i] <- x
    } else {
      idxBefore <- max(which(V < x), na.rm = TRUE)
      idxAfter  <- min(which(V > x), na.rm = TRUE)
      if (length(idxBefore) == 0) idxBefore <- idxAfter
      if (length(idxAfter)  == 0) idxAfter  <- idxBefore
      if (abs(V[idxBefore] - x) < abs(V[idxAfter] - x)) {
        results[i] <- V[idxBefore]
      } else if (abs(V[idxBefore] - x) > abs(V[idxAfter] - x)) {
        results[i] <- V[idxAfter]
      } else {
        results[i] <- mean(c(V[idxBefore], V[idxAfter]))
      }
    }
  }
  results
}

###############################################################################
# invertDnDraws.R — Direct kmax inversion from Dn realisations
###############################################################################

#' Invert Dn draws to kmax(Da)
#'
#' For each realisation s, inverts each model's Dn(ky) curve to kmax(Da),
#' builds weighted ensemble mean, then averages over realisations.
#'
#' Log-log linear extrapolation is used when Da falls outside the support of
#' the computed Dn(ky) curve. No calibration-range clamping is applied — the
#' natural support of the curve bounds the inversion.
#'
#' All inputs must be in the same displacement units (cm).
#'
#' @param draws   data.table(ky, s, Dn, IDn) — Dn in cm
#' @param Da      numeric vector — displacement targets in cm
#' @param weights named numeric vector, ensemble weights by IDn
#' @param p       numeric quantiles in (0,1) to report alongside the mean.
#'                Default c(0.05, 0.10, 0.16, 0.84, 0.90, 0.95). The output
#'                always includes the ensemble mean as the first p row.
#'
#' @return data.table(Da, p, kmax). Column `p` is character: "mean" or the
#'         numeric quantile formatted as "0.84", "0.90", etc.
#' @examples
#' \dontrun{
#' uhs <- data.table::fread(
#'   system.file("extdata", "uhs.csv", package = "newmark"))
#' out <- fitDnCurve(uhs, ky = getDnKy(uhs, Ts = 0.3), Ts = 0.3, Mw = 7.5,
#'                   NS = 100,
#'                   weights = c(AM88=1, BT07=1, SR08=1, BM17=1))
#' invertDnDraws(out$draws, Da = c(2.5, 5, 10),
#'               weights = c(AM88=1, BT07=1, SR08=1, BM17=1))
#' }
#' @export
invertDnDraws <- function(draws, Da, weights,
                          p = c(0.05, 0.10, 0.16, 0.84, 0.90, 0.95)) {

  stopifnot(is.numeric(p), all(p > 0 & p < 1))

  w   <- as.numeric(weights)
  w   <- w / sum(w)
  ids <- names(weights)
  S   <- sort(unique(draws$s))

  OUT <- vector("list", length(S))

  for (j in seq_along(S)) {
    si <- S[j]
    km <- numeric(length(Da))

    for (k in seq_along(Da)) {
      x <- numeric(length(ids))
      for (i in seq_along(ids)) {
        DT <- draws[s == si & IDn == ids[i]]
        if (!nrow(DT)) {
          x[i] <- NA_real_
          next
        }
        data.table::setorder(DT, ky)
        # force Dn non-increasing in ky (monotone projection)
        DT[, Dn := rev(cummax(rev(Dn)))]
        # invert in log-log: Dn decreasing -> ky increasing
        # approx needs x increasing, so use Dn reversed
        lDn <- rev(log(DT$Dn + .Machine$double.eps))
        lky <- rev(log(DT$ky))
        lDa <- log(Da[k])
        n   <- length(lDn)
        if (lDa >= lDn[1L] && lDa <= lDn[n]) {
          x[i] <- exp(stats::approx(lDn, lky, xout = lDa,
                                    rule = 1L, ties = "ordered")$y)
        } else if (lDa > lDn[n]) {
          t <- (lky[n] - lky[n - 1L]) / (lDn[n] - lDn[n - 1L])
          x[i] <- exp(lky[n] + t * (lDa - lDn[n]))
        } else {
          t <- (lky[2L] - lky[1L]) / (lDn[2L] - lDn[1L])
          x[i] <- exp(lky[1L] + t * (lDa - lDn[1L]))
        }
      }
      km[k] <- weighted.mean(x, w, na.rm = TRUE)
    }

    OUT[[j]] <- data.table::data.table(Da = Da, kmax = km)
  }

  DT <- data.table::rbindlist(OUT, use.names = TRUE)
  DT[, list(
    p    = c("mean", sprintf("%.2f", p)),
    kmax = c(mean(kmax, na.rm = TRUE),
             stats::quantile(kmax, p, na.rm = TRUE, names = FALSE))
  ), by = .(Da)]
}

# nolint start
#' Build Ground-Motion Design Parameters (GMDP)
#'
#' Assembles an annual-exceedance-probability (AEP) table, uniform-hazard-spectrum
#' (UHS) table, and--where available--a magnitude-distance disaggregation table,
#' using OpenQuake or user-supplied hazard files.
#'
#' @param IDo    Character. Identifier for the GMDP. Default `"gmdp"`.
#' @param path   Character. Directory with hazard data files.
#' @param engine Character. `"openquake"` (default) or `"user"`.
#' @param vref   Numeric. Reference Vs30 (m/s). Default 760.
#' @param TRo    Numeric vector. Target return periods (years).
#'
#' @import data.table

#' @return A named list with `AEPTable`, `UHSTable`, and `RMwTable`
#'   (`NULL` if disaggregation is unavailable).
#'
#' @examples
#' \dontrun{
#' # Requires an 'OpenQuake' Engine output directory containing classical
#' # PSHA + disaggregation ZIPs; not bundled with the package.
#' out <- buildGMDP(path = "path/to/openquake/output", vref = 760)
#' }
#' @export
buildGMDP <- function(path,
                      IDo = "GEM",
                      engine = "openquake",
                      vref = 760,
                      TRo = c(100, 200, 500, 1000, 2000, 2500, 5000, 10000)) {
    OUT <- importAEPTable(path = path, engine = engine, vref = vref)
    AEPTable <- OUT$AEPTable
    ITo <- unique(AEPTable$ITo)[1]
    RMwTable <- NULL

    if (engine == "openquake") {
        message("> Building Disaggregation Hazard Table...")
        RMwTable <- tryCatch(
            importModel.oqRMw(path = path, ITo = ITo, vref = vref),
            error = function(e) {
                message(">> Skipping disagg (", e$message, ")")
                NULL
            }
        )
        if (!is.null(RMwTable)) {
            RMwTable <- RMwTable[, `:=`(SID = Vs30toSID(vref), Vs30 = vref, SM = engine, IT = ITo)]
        } else {
            message("> Disaggregation data not available.")
        }
    }

    message("> Re-mesh hazard data on a uniform TR grid ...")
    COLS <- setdiff(names(AEPTable), c("Sa", "POE", "AEP", "TR"))
    # Include all output kinds (mean, std, quantiles) as separate p-levels.
    # This ensures downstream stages can propagate non-mean curves without breaking.
    UHSTable <- AEPTable[Tn != -1, remeshGroup(.SD, TRo, .BY), by = COLS]

    UHSTable[, `:=`(Vref = vref, Vs30 = vref, AF = 1, SaF = Sa)]
    UHSTable[, ID := IDo]
    AEPTable[, ID := IDo]
    if (!any(UHSTable$Tn == 0)) {
        stop("Missing PGA (Tn==0). Check openQuake job.ini file.")
    }
    list(
        AEPTable = AEPTable,
        UHSTable = UHSTable,
        RMwTable = RMwTable
    )
}

#' Import OpenQuake or user AEP Data (Internal)
#'
#' @keywords internal
#' @noRd
importAEPTable <- function(path, engine = "openquake", vref = 760) {
    message("> Build AEP Table...")
    AEPTable <- NULL

    if (engine == "openquake") {
        message("> Unzip OQ data ...")

        FILES <- list.files(path, pattern = "\\.zip$", full.names = TRUE)
        if (!length(FILES)) {
            stop("No .zip files found in ", path)
        }

        for (z in FILES) {
            message("> Import AEP data from: ", z)
            tryCatch(
                {
                    TempDir <- tempdir()
                    unlink(TempDir, recursive = TRUE)
                    dir.create(TempDir, showWarnings = FALSE)
                    utils::unzip(z, junkpaths = TRUE, exdir = TempDir)

                    AUX <- importModel.oqAEP(path = TempDir, vref = vref)
                    AEPTable <- if (is.null(AEPTable)) {
                        AUX
                    } else {
                        data.table::rbindlist(list(AEPTable, AUX), use.names = TRUE, fill = TRUE)
                    }
                },
                error = function(e) {
                    message(">> Skipping file: ", z, " (", e$message, ")")
                }
            )
        }
    } else if (engine == "user") {
        message("> Import user-supplied AEP.xlsx ...")
        AEPTable <- tryCatch(
            importModel.userAEP(path, filename = "AEP.xlsx"),
            error = function(e) stop("Error in user AEP data: ", e$message)
        )
    } else {
        stop("Unknown engine: ", engine)
    }

    if (is.null(AEPTable) || !nrow(AEPTable)) {
        stop("No valid AEP data found in path: ", path)
    }
    AEPTable[, `:=`(Vref = vref, Vs30 = vref)]
    return(list(AEPTable = AEPTable[]))
}

#' Import OpenQuake AEP File (Internal)
#'
#' @keywords internal
#' @noRd
importModel.oqAEP <- function(path, vref) {
    DT <- data.table::data.table()

    FILES <- c(
        list.files(path, pattern = "quantile_curve"),
        list.files(path, pattern = "hazard_curve"),
        list.files(path, pattern = "hcurves-csv")
    )
    if (!length(FILES)) {
        stop("No hazard_curve, quantile_curve, or hcurves-csv files found in ", path)
    }

    for (.files in FILES) {
        HEADER <- readLines(file.path(path, .files), n = 1)
        p <- .extractQuantileFromHeader(HEADER)
        ITo <- .extractInvestigationTime(HEADER)
        Tn <- .extractTnFromHeader(HEADER)

        if (is.na(p) || is.na(ITo)) {
            stop("Malformed header in file: ", file.path(path, .files))
        }

        AUX <- data.table::fread(file.path(path, .files), skip = 1, header = FALSE, blank.lines.skip = TRUE)
        if (!nrow(AUX)) {
            warning("Empty data in file: ", file.path(path, .files))
            next
        }

        COLS <- unlist(AUX[1]) |> as.vector()
        data.table::setnames(AUX, COLS)
        AUX <- AUX[-1]

        poeCols <- grep("poe-", COLS, value = TRUE)
        if (!length(poeCols)) {
            stop("No 'poe-' columns in file: ", file.path(path, .files))
        }

        DATA <- data.table::melt(
            data = AUX,
            id.vars = c("lon", "lat", "depth"),
            measure.vars = poeCols,
            variable.name = "Sa",
            value.name = "POE"
        )
        DATA[, POE := as.numeric(POE)]
        DATA[, Sa := as.numeric(sub("^poe-", "", Sa))]
        DATA[, AEP := poeToAep(POE, ITo)]
        DATA[, TR := poeToTr(POE, ITo)]
        DATA[, `:=`(Tn = Tn, p = p, ITo = ITo)]

        DT <- data.table::rbindlist(list(DT, DATA), use.names = TRUE)
    }

    # remove infinite or negative TR
    DT <- DT[is.finite(TR) & TR > 0 & !is.na(TR)]

    if (!nrow(DT)) {
        stop("No valid hazard data in: ", path)
    }
    ## ---- Check: debe existir PGA ----------------------------------------
    if (!any(DT$Tn == 0)) {
        stop("PGA row not found after import; Check input files. Check job.ini.")
    }

    return(DT[])
}

#' Import OpenQuake Mag-Dist Disaggregation (Internal)
#'
#' @keywords internal
#' @noRd
importModel.oqRMw <- function(path, ITo, vref) {
    if (!dir.exists(path)) {
        stop("Path does not exist: ", path)
    }

    DIR <- file.path(path, paste0(".temp_oqRMw_", as.integer(Sys.time())))
    if (dir.exists(DIR)) {
        unlink(DIR, recursive = TRUE, force = TRUE)
    }
    dir.create(DIR, showWarnings = FALSE)

    FILES <- list.files(path, pattern = "\\.zip$", full.names = TRUE)
    if (length(FILES)) {
        for (.files in FILES) {
            utils::unzip(.files, exdir = DIR, junkpaths = TRUE)
        }
    }
    searchDir <- if (length(FILES)) DIR else path

    FILES <- list.files(searchDir, pattern = "Mag_Dist", full.names = TRUE)
    FILES <- FILES[!grepl("TRT", FILES)]
    if (!length(FILES)) {
        unlink(DIR, recursive = TRUE, force = TRUE)
        message("> No 'Mag_Dist' files found in path. Skipping...")
        return(NULL)
    }
    meanFiles <- grep("Mag_Dist-mean", FILES, value = TRUE)
    if (length(meanFiles)) {
        FILES <- meanFiles
    }

    DHT <- data.table::data.table()
    for (.files in FILES) {
        DT <- tryCatch(
            data.table::fread(.files, skip = 1, header = TRUE, blank.lines.skip = TRUE),
            error = function(e) NULL
        )
        if (is.null(DT) || !nrow(DT)) {
            next
        }

        requiredCols <- c("mag", "dist", "poe", "imt")
        missingCols <- setdiff(requiredCols, names(DT))
        if (length(missingCols)) {
            next
        }

        data.table::setnames(DT, old = c("mag", "dist", "poe"), new = c("Mw", "R", "POE"), skip_absent = TRUE)
        if ("iml" %in% names(DT)) {
            DT[, iml := NULL]
        }

        rlzCol <- grep("rlz|mean", names(DT), value = TRUE)
        if (length(rlzCol) == 1) {
            data.table::setnames(DT, old = rlzCol, new = "p")
        } else {
            next
        }

        DT[imt == "PGA", imt := "Sa(0.0)"]
        DT[, Tn := stringr::str_extract(imt, "(?<=\\()\\d+\\.*\\d*(?=\\))")]
        DT[, Tn := as.numeric(Tn)]
        DT[is.na(Tn), Tn := 0]
        DT[, IT := ITo]
        DT[, `:=`(AEP = poeToAep(POE, IT), TR = poeToTr(POE, IT))]
        DT <- DT[is.finite(TR) & TR > 0 & is.finite(AEP) & AEP > 0]

        DHT <- data.table::rbindlist(list(DHT, DT), fill = TRUE)
    }

    unlink(DIR, recursive = TRUE, force = TRUE)
    if (!nrow(DHT)) {
        message("No valid disagg data found.")
        return(NULL)
    }
    return(unique(DHT))
}

#' Internal: Extract Quantile from OpenQuake Header
#' @keywords internal
#' @noRd
.extractQuantileFromHeader <- function(line) {
    if (grepl("kind='mean'", line)) {
        return("mean")
    }
    if (grepl("kind='std'", line)) {
        # Standard deviation curve exported by OpenQuake
        return("std")
    }
    mt <- regexpr("kind='quantile-([0-9\\.]+)'", line)
    if (mt > 0) {
        val <- regmatches(line, mt)
        qval <- sub("kind='quantile-", "", val)
        qval <- sub("'", "", qval)
        return(as.numeric(qval))
    }
    return(NA)
}



#' @keywords internal
#' @noRd
.extractInvestigationTime <- function(line) {
    mt <- regexpr("investigation_time=([0-9\\.]+)", line)
    if (mt > 0) {
        val <- regmatches(line, mt)
        num <- sub("investigation_time=", "", val)
        return(as.numeric(num))
    }
    return(NA_real_)
}

#' @keywords internal
#' @noRd
.extractTnFromHeader <- function(line) {
    l <- tolower(line) # ignora mayúsculas
    # --- PGA --------------------------------------------------------------
    if (grepl("imt=[\"']?pga[\"']?", l)) {
        return(0)
    }
    # --- PGV --------------------------------------------------------------
    if (grepl("imt=[\"']?pgv[\"']?", l)) {
        return(-1)
    }
    # --- SA(T) ------------------------------------------------------------
    mt <- regexpr("imt=[\"']?sa\\(([0-9\\.]+)\\)[\"']?", l)
    if (mt > 0) {
        val <- regmatches(l, mt)
        num <- sub("imt=[\"']?sa\\(", "", val)
        num <- sub("\\)[\"']?", "", num)
        return(as.numeric(num))
    }
    return(NA_real_)
}

#' Internal: Interpolate Hazard Curve to Uniform TR Grid
#' @keywords internal
#' @noRd
.formatRemeshContext <- function(groupBy = NULL) {
    if (is.null(groupBy) || !length(groupBy)) {
        return("hazard curve")
    }
    AUX <- vapply(names(groupBy), function(x) paste0(x, "=", as.character(groupBy[[x]])), character(1))
    paste(AUX, collapse = ", ")
}

.prepareCurveForRemesh <- function(TRi, Sai, groupBy = NULL) {
    DT <- data.table::data.table(TR = as.numeric(TRi), Sa = as.numeric(Sai))
    DT <- DT[is.finite(TR) & TR > 0 & is.finite(Sa) & Sa > 0]
    if (nrow(DT) < 2L) {
        stop("Need at least two positive (TR, Sa) points for ", .formatRemeshContext(groupBy), ".")
    }

    data.table::setorder(DT, TR)
    if (anyDuplicated(DT$TR)) {
        AUX <- DT[, .(
            Sa = max(Sa),
            nSa = data.table::uniqueN(signif(Sa, 12))
        ), by = TR]
        if (any(AUX$nSa > 1L)) {
            message(">> Collapsing duplicate TR nodes by keeping max Sa in ", .formatRemeshContext(groupBy), ".")
        }
        DT <- AUX[, .(TR, Sa)]
        data.table::setorder(DT, TR)
    }
    if (nrow(DT) < 2L) {
        stop("Need at least two unique TR nodes for ", .formatRemeshContext(groupBy), ".")
    }
    DT
}

.checkTrCoverage <- function(TRi, TRo, groupBy = NULL) {
    trMin <- min(TRi)
    trMax <- max(TRi)
    targetMin <- min(TRo)
    targetMax <- max(TRo)
    tol <- max(0.5, 1e-4 * max(c(abs(TRi), abs(TRo))))
    if (targetMin < trMin - tol || targetMax > trMax + tol) {
        stop(
            "Requested TR grid [", signif(targetMin, 8), ", ", signif(targetMax, 8),
            "] exceeds imported curve range [", signif(trMin, 8), ", ", signif(trMax, 8),
            "] for ", .formatRemeshContext(groupBy),
            ". Expand the oqt POE grid or narrow TRo instead of extrapolating."
        )
    }
    tol
}

reMeshCurve <- function(TRi, Sai, TRo, groupBy = NULL) {
    if (!length(TRo) || any(!is.finite(TRo) | TRo <= 0)) {
        stop("TRo must contain finite positive return periods.")
    }

    DT <- .prepareCurveForRemesh(TRi, Sai, groupBy)
    tol <- .checkTrCoverage(DT$TR, TRo, groupBy)
    TRfit <- TRo
    TRfit[TRfit < min(DT$TR)] <- min(DT$TR)
    TRfit[TRfit > max(DT$TR)] <- max(DT$TR)
    if (any(TRo < min(DT$TR) - tol | TRo > max(DT$TR) + tol)) {
        stop("Requested TR grid exceeds the tolerated interpolation support.")
    }

    logTarget <- stats::approx(
        x = log(DT$TR),
        y = log(DT$Sa),
        xout = log(TRfit),
        method = "linear",
        ties = "ordered"
    )$y
    if (any(!is.finite(logTarget))) {
        stop("Interpolation failed for ", .formatRemeshContext(groupBy), ".")
    }
    exp(logTarget)
}


#' Internal: Remesh a Group to Uniform TR Grid
#' @keywords internal
#' @noRd
remeshGroup <- function(.SD, TRo, groupBy = NULL) {
    if (!all(c("TR", "Sa") %in% colnames(.SD))) {
        stop("'.SD' must contain columns 'TR' and 'Sa'.")
    }

    data.table::setorder(.SD, TR)
    .SD <- .SD[is.finite(TR) & TR > 0]
    if (nrow(.SD) < 2) {
        return(NULL)
    }

    TRi <- .SD$TR
    Sai <- .SD$Sa
    SaStar <- reMeshCurve(TRi, Sai, TRo, groupBy)
    data.table(TR = TRo, Sa = SaStar)
}

#' Internal: Remesh a Group to Uniform TR Grid
#' @keywords internal
#' @noRd
importModel.userAEP <- function(path = NULL, filename = "AEP.xlsx") {
    if (is.null(path) || !dir.exists(path)) {
        stop("Invalid `path`: must be a directory.")
    }

    inputFile <- file.path(path, filename)
    if (!file.exists(inputFile)) {
        stop("File not found: ", inputFile)
    }

    sheetsAll <- readxl::excel_sheets(inputFile)
    sheetsP <- grep(pattern = "^p=", sheetsAll, value = TRUE)
    if (length(sheetsP) == 0) {
        stop("No sheets named 'p=...' found in: ", inputFile)
    }

    AT <- data.table()
    for (SHEET in sheetsP) {
        DT <- data.table::as.data.table(readxl::read_xlsx(inputFile, sheet = SHEET))
        if (!("Tn" %in% names(DT))) {
            stop("Missing column 'Tn' in sheet '", SHEET, "'.")
        }
        idVar <- "Tn"
        measureVars <- setdiff(names(DT), idVar)
        if (length(measureVars) == 0) {
            warning("No measure columns in sheet '", SHEET, "'. Skipping.")
            next
        }

        aux <- melt(
            DT,
            id.vars = idVar,
            measure.vars = measureVars,
            variable.name = "Sa",
            value.name = "AEP",
            variable.factor = FALSE
        )
        aux[, Sa := as.character(Sa)]

        po <- stringr::str_remove(SHEET, "p=")
        if (po != "mean") {
            poNum <- suppressWarnings(as.numeric(po))
            if (!is.na(poNum)) {
                po <- poNum
            }
        }

        aux <- aux[AEP > 0]
        if (!is.numeric(aux$Sa)) {
            aux[, Sa := suppressWarnings(as.numeric(as.character(Sa)))]
        }

        aux[, p := po]
        aux[, IT := 50]
        aux[, POE := aepToPoe(AEP, IT)]
        aux[, TR := aepToTr(AEP)]

        AT <- rbind(AT, aux, use.names = TRUE, fill = TRUE)
    }

    colOrder <- c("Tn", "Sa", "AEP", "p", "POE", "TR", "IT")
    colOrder <- intersect(colOrder, names(AT))
    setcolorder(AT, colOrder)

    return(AT[])
}

#' Baker and Jayaram (2008) inter-period correlation model
#'
#' Computes the Pearson correlation coefficient between the residuals of
#' ln(Sa) at two oscillator periods, using the canonical predictive equation
#' of Baker and Jayaram (2008), Earthquake Spectra 24(1), 299-317
#' (DOI 10.1193/1.2857544), Eqs. 5 and 6 verbatim.
#'
#' The model has no dependence on rupture distance or magnitude.
#' \code{Rrup} is retained for signature compatibility and is silently ignored.
#' Periods of zero are treated as 0.01 s (standard PGA proxy in B&J 2008).
#'
#' Form (Eqs. 5-6 of B&J 2008):
#'   C1 = 1 - cos(pi/2 - 0.366 * ln(Tmax / max(Tmin, 0.109)))
#'   C2 = (1 - 0.105 * (1 - 1/(1 + exp(100*Tmax - 5))) * (Tmax-Tmin)/(Tmax-0.0099))
#'         if Tmax < 0.2; else 0
#'   C3 = C2 if Tmax < 0.109; else C1
#'   C4 = C1 + 0.5 * (sqrt(C3) - C3) * (1 + cos(pi * Tmin / 0.109))
#'
#'   rho = C2          if Tmax < 0.109
#'       = C1          else if Tmin > 0.109
#'       = min(C2, C4) else if Tmax < 0.2
#'       = C4          otherwise
#'
#' Validity: 0.01 <= T1, T2 <= 10 s.
#'
#' @param Tn   Numeric scalar. Target spectral period in seconds.
#' @param T0   Numeric scalar. Reference period in seconds. Default 0.01
#'   (used as a numerical proxy for PGA). Values <= 0 are silently
#'   converted to 0.01.
#' @param Rrup Numeric scalar. Not used; retained for backward compatibility.
#'
#' @return Numeric scalar correlation coefficient rho(T0, Tn).
#'
#' @examples
#' rhoBJ(Tn = 0.3, T0 = 0.01)   # rho(PGA, 0.3 s)
#' rhoBJ(Tn = 1.0, T0 = 0.5)
#'
#' @references
#' Baker, J. W. & Jayaram, N. (2008). Correlation of Spectral Acceleration
#' Values from NGA Ground Motion Models. Earthquake Spectra, 24(1), 299-317.
#' DOI: 10.1193/1.2857544.
#'
#' @export
rhoBJ <- function(Tn, T0 = 0.01, Rrup = 100) {
  if (T0 <= 0) T0 <- 0.01
  if (Tn <= 0) Tn <- 0.01
  stopifnot(length(Tn) == 1L, length(T0) == 1L)
  stopifnot(is.finite(Tn), is.finite(T0))

  if (any(c(T0, Tn) < 0.01) || any(c(T0, Tn) > 10)) {
    warning("rhoBJ: B&J 2008 is calibrated for T in [0.01, 10] s. ",
            "Extrapolation outside this range is not supported by the source paper.")
  }

  Tmin <- min(T0, Tn)
  Tmax <- max(T0, Tn)

  C1 <- 1 - cos(pi / 2 - 0.366 * log(Tmax / max(Tmin, 0.109)))

  if (Tmax < 0.2) {
    C2 <- 1 - 0.105 * (1 - 1 / (1 + exp(100 * Tmax - 5))) *
      (Tmax - Tmin) / (Tmax - 0.0099)
  } else {
    C2 <- 0
  }

  C3 <- if (Tmax < 0.109) C2 else C1
  C4 <- C1 + 0.5 * (sqrt(C3) - C3) * (1 + cos(pi * Tmin / 0.109))

  if (Tmax < 0.109) {
    rho <- C2
  } else if (Tmin > 0.109) {
    rho <- C1
  } else if (Tmax < 0.2) {
    rho <- min(C2, C4)
  } else {
    rho <- C4
  }

  rho
}

#' Draw a correlated sample of spectral accelerations
#'
#' Draw a correlated sample of spectral accelerations
#'
#' @param uhs data.table with columns Tn, p, Sa (must include p == "mean").
#' @param Tn  numeric vector; the first element is the reference period \code{Tn[1]},
#'            e.g., 0.01 or 0.00. It does not have to be sorted.
#' @param rho numeric vector of length \code{length(Tn) - 1}; correlations between the
#'            reference period \code{Tn[1]} and each remaining period.
#' @param NS  integer, Monte Carlo sample size.
#' @return matrix of dimension NS x length(Tn) with Sa draws (g).
#' @importFrom mvtnorm rmvnorm
#' @importFrom stats    pnorm
#' @examples
#' \dontrun{
#' uhs <- data.table::data.table(
#'   Tn = c(0, 0, 0, 1, 1, 1),
#'   p  = c("0.16", "mean", "0.84", "0.16", "mean", "0.84"),
#'   Sa = c(0.3, 0.5, 0.7, 0.1, 0.2, 0.3)
#' )
#' rho <- rhoBJ(Tn = 1.0, T0 = 0.01)
#' mat <- sampleSaCorr(uhs, Tn = c(0, 1), rho = rho, NS = 100)
#' }
#' @export

sampleSaCorr <- function(uhs, Tn, rho, NS) {
  if (length(rho) != length(Tn) - 1L) {
    stop("length(rho) must be length(Tn) - 1")
  }

  ## 1) Correlation matrix (reference period is Tn[1])
  C <- diag(length(Tn))
  C[1, -1] <- C[-1, 1] <- as.numeric(rho)

  .sampleFromCorMatrix(C, uhs, Tn, NS)
}


.sampleSaCorrFull <- function(uhs, Tn, rho, NS) {
  stopifnot(is.matrix(rho), nrow(rho) == length(Tn), ncol(rho) == length(Tn))
  .sampleFromCorMatrix(rho, uhs, Tn, NS)
}


# Shared body of sampleSaCorr() and .sampleSaCorrFull(): given a fully
# constructed correlation matrix C, draw NS correlated Sa samples per
# Tn via Gaussian copula + empirical inverse + recentering to tabulated
# mean. Both wrappers above only differ in how they build C.
.sampleFromCorMatrix <- function(C, uhs, Tn, NS) {
  ## Ensure positive-definite numerically
  ev <- eigen(C, symmetric = TRUE, only.values = TRUE)$values
  if (min(ev) <= 0) {
    C <- C + diag(abs(min(ev)) + 1e-8, nrow(C))
  }

  ## Gaussian copula -> uniforms
  Z <- mvtnorm::rmvnorm(NS, sigma = C)
  U <- stats::pnorm(Z)

  ## Map to Sa via empirical inverse; re-center to tabulated mean at each Tn
  out <- matrix(NA_real_, NS, length(Tn))
  for (j in seq_along(Tn)) {
    Dt <- uhs[Tn == Tn[j] & p != "mean", .(p, Sa)]
    if (nrow(Dt) < 2L) stop("Not enough quantiles at Tn = ", Tn[j])
    Dt[, p := as.numeric(p)]

    Qln <- buildQSpline(Dt)
    Q   <- exp(Qln(U[, j]))

    mu <- uhs[Tn == Tn[j] & p == "mean", Sa][1]
    if (!is.finite(mu)) mu <- mean(Q)
    out[, j] <- Q * (mu / mean(Q))
  }

  out
}

# nolint start
#' Single-realisation synthetic soil profile and site properties
#'
#' Builds one Monte Carlo realisation of a layered soil column from a
#' total height \code{Hs} and a USCS classification, samples random
#' geotechnical properties (void ratio, unit weight, plasticity,
#' shear modulus parameters), computes the fundamental period
#' \code{Ts} via Rayleigh's method (\code{fitModel.Ts}) and the
#' Vs30-equivalent shear-wave velocity. \code{getSiteProperties}
#' calls this function repeatedly to assemble the Monte Carlo
#' ensemble; call directly when one realisation suffices.
#'
#' @param Hs Numeric scalar. Total height to hard ground (m).
#' @param Water Numeric scalar. Water-table depth as fraction of
#'   \code{Hs}: \code{zw = Hs - Water * Hs}.
#' @param USCS Character vector. Unified Soil Classification System
#'   codes (e.g. \code{c("GC", "CL", "ML")}).
#' @param Group Character vector or \code{NULL}. Soil groups,
#'   \code{c("Gravels","Fines","Sands")}; if \code{NULL}, derived from
#'   \code{USCS}.
#' @param h Numeric scalar. Layer thickness in the discretisation
#'   (m). Default 0.50.
#' @param DrID Character or \code{NULL}. Relative-density label,
#'   \code{c("Very Dense","Dense","Compact","Loose","Very Loose")};
#'   \code{NULL} samples relative density from the standard prior.
#' @param UniformDistribution Logical. If \code{TRUE} (default),
#'   sample soil properties from uniform distributions; if
#'   \code{FALSE}, use normal distributions.
#' @param POP Numeric. Pre-consolidation pressure (kPa); enters the
#'   over-consolidation ratio \code{OCR = (pm + POP) / pm}.
#'   Default 100.
#' @param IgnoreModelIntervals Logical. If \code{TRUE} (default),
#'   ignore model-specific layer-thickness intervals and use \code{h}
#'   as the discretisation step.
#' @param getSiteLayers Logical. If \code{TRUE}, return the per-layer
#'   detail in addition to the per-site summary.
#' @param Vref Numeric. Reference shear-wave velocity (m/s) for the
#'   site-class assignment. Default 760.
#'
#' @return data.table with one row per (Hs, POP, Hw, sample)
#'   combination and columns Hs, Hw, NL, Z500, Z1000, SID, Go, mo,
#'   Ts, VSo, VS30, plus USCS-fraction columns.
#' @export geSiteTable
#'
#' @import data.table
#' @import digest
#' @import stats
#' @importFrom triangle rtriangle
#' @import stringr
#' @import utils
#'
geSiteTable <- function(Hs, Water = 0, USCS, Group = NULL, h = 0.50, DrID = NULL, Vref = 760, UniformDistribution = TRUE, POP = 100, IgnoreModelIntervals = TRUE, getSiteLayers = FALSE) {
    . <- NULL

    # Ensure the lazy-loaded dataset is bound in this scope. Without
    # this call, bare-name references below resolve via the package
    # namespace under `library(newmark)` but can fail to resolve when
    # the package is loaded via `pkgload::load_all()` (the path pkgdown
    # takes when install = FALSE), producing
    #   "object 'ShearModelParameters' not found".
    utils::data("ShearModelParameters", package = "newmark",
                envir = environment())



    NL <- ceiling(Hs / h)
    # Dimensions  ----
    OV <- vector(mode = "double", length = NL)
    OS <- vector(mode = "character", length = NL)
    # Set vertical coordinates ----
    hs <- rep(h, NL)
    zo <- OV
    zi <- OV
    zm <- OV
    for (k in seq(1, NL)) {
        if (k > 1) {
            zo[k] <- zi[k - 1]
        }
        zi[k] <- zo[k] + h
        zm[k] <- (zi[k] + zo[k]) / 2
    }

    # Set group ID ----
    if (is.null(Group) & is.null(USCS)) {
        USCS <- ValidUSCS
    }

    # Case 2: Sampling soils from the same gropu(s)

    if (!is.null(Group) & is.null(USCS)) {
        OK <- tolower(Group) %in% c("gravels", "sands", "fines", "silts", "clays", "organic")
        stopifnot(OK)
        USCS <- switch(tolower(Group),
            gravels = ValidGravels,
            sands = ValidSands,
            fines = ValidFines,
            silts = ValidSilts,
            clays = ValidClays,
            silts = ValidSilts
        )
    }

    stopifnot(!is.null(USCS) && all(USCS %in% ValidUSCS))
    UID <- sample(USCS, size = NL, replace = TRUE)
    GID <- OV
    gs <- OV # Sumerged,Saturated, depending on zw
    gd <- OV # Dry
    gsat <- OV # Sat
    eo <- OV
    emax <- OV
    emin <- OV
    gsmax <- OV
    gsmin <- OV
    Dr <- OV
    LL <- OV
    IP <- OV
    po <- OV
    pi <- OV
    pm <- OV
    Gm <- OV
    VSm <- OV
    OCR <- OV
    gw <- OV
    # rho <- OV
    Hw <- Hs * Water

    zw <- Hs - Hw

    for (k in seq(1, NL)) {
        # Groupid
        if (UID[k] %in% c(ValidGravels)) {
            GID[k] <- "Gravels"
        }
        if (UID[k] %in% ValidSands) {
            GID[k] <- "Sands"
        }
        if (UID[k] %in% ValidFines) {
            GID[k] <- "Fines"
        }

        # Void Ratios ----
        if (is.null(DrID)) {
            RANGE <- VoidRatiosUSCS[USCS == UID[k]]
        }
        if (!is.null(DrID)) {
            RANGE <- RelativeDensityRanges[toupper(DensityID) == toupper(DrID) & GroupID == GID[k]]
        }

        #
        if (IgnoreModelIntervals == TRUE) {
            emin[k] <- runif(1, min = RANGE$eminMin, max = RANGE$eminMax)
            emax[k] <- runif(1, min = RANGE$emaxMin, max = RANGE$emaxMax)

            if (UniformDistribution == TRUE) {
                eo[k] <- runif(1, min = emin[k], max = emax[k])
            } else {
                eo[k] <- rtriangle(1, a = emin[k], b = emax[k], c = 1 / 2 * (emin[k] + emax[k]))
            }
        }

        if (IgnoreModelIntervals == FALSE) {
            # Build ranges within the model ranges.
            AvailableModels <- FALSE
            NIT <- 0
            NITmax <- 100
            while (AvailableModels == FALSE & NIT < NITmax) {
                NIT <- NIT + 1
                emin[k] <- runif(1, min = RANGE$eminMin, max = RANGE$eminMax)
                emax[k] <- runif(1, min = RANGE$emaxMin, max = RANGE$emaxMax)

                # No risk of emin==emax since ranges were splitted

                if (UniformDistribution == TRUE) {
                    eo[k] <- runif(1, min = emin[k], max = emax[k])
                } else {
                    eo[k] <- rtriangle(1, a = emin[k], b = emax[k], c = 1 / 2 * (emin[k] + emax[k]))
                }

                # Shear Module ----
                MID <- ShearModelParameters[GroupID == GID[k]][eo[k] >= emin & eo[k] <= emax]$ModelID
                if (length(MID) > 0) {
                    AvailableModels <- TRUE
                } else {
                    AvailableModels <- FALSE
                }
            }
            if (NIT >= NITmax) {
                return(NULL)
            }
        }


        # Relative Density
        Dr[k] <- (emax[k] - eo[k]) / (emax[k] - emin[k])
        # Unit Weight kN/m3
        RANGE <- UnitWeightRanges[USCS == UID[k]]

        # **********************************
        # gsmin[k] <- runif(1,min=RANGE$gsminMin,max=RANGE$gsminMax) #[kN/m3]
        # gsmax[k] <- runif(1,min=RANGE$gsmaxMin,max=RANGE$gsmaxMax) #[kN/m3]
        gsmin[k] <- RANGE$gsatMin # [kN/m3]
        gsmax[k] <- RANGE$gsatMax # [kN/m3]
        # **********************************
        # Saturated wunit weight
        # gsat[k] <- gsmax[k] - (gsmax[k] - gsmin[k]) * (1 - Dr[k]) #[kN/m3]
        gsat[k] <- runif(n = 1, min = gsmin[k], max = gsmax[k])

        # Dry weight
        gd[k] <- max(0, gsat[k] - 10)



        # Submerged weight
        if (zm[k] > zw) {
            gw[k] <- 10
            gs[k] <- max(0, gsat[k] - gw[k]) # some organic clays has gs<10
        } else {
            gs[k] <- gsat[k]
        }
        # Density
        # rho[k] <- gs[k]*1000/9.81 # gs in KN/m3, rho in kg/m3


        # Plasticity Index

        if (UID[k] %in% ValidFines) {
            if (UID[k] %in% c("CH", "MH", "OH")) {
                LL[k] <- runif(n = 1, min = 50, max = 100)
            }
            if (UID[k] %in% c("CL", "ML", "OL")) {
                LL[k] <- runif(n = 1, min = 8, max = 50)
            }
            U_LINE <- 0.9 * (LL[k] - 8) # >8 by def LL
            A_LINE <- ifelse(LL[k] >= 20, 0.73 * (LL[k] - 20), 0)
            if (UID[k] %in% ValidClays) {
                IP[k] <- runif(n = 1, min = A_LINE, max = U_LINE)
            }
            if (UID[k] %in% c(ValidSilts, ValidOrganic)) {
                IP[k] <- runif(n = 1, min = 0, max = A_LINE)
            }
        }

        # Octahedral pressures
        Ko <- 0.5
        if (k > 1) {
            po[k] <- pi[k - 1]
        }


        Dp <- 1 / 3 * (1 + 2 * Ko) * gs[k] * hs[k] # [kN/m3->kPa]
        pi[k] <- po[k] + Dp # kPa
        pm[k] <- (pi[k] + po[k]) / 2 # kPa

        # OverConsolidation Ratio
        OCR[k] <- max((pm[k] + POP) / pm[k], 1)


        # Shear Module ----
        MID <- ShearModelParameters[GroupID == GID[k]]$ModelID
        A <- ShearModelParameters[ModelID %in% MID, A] # kPa
        Ce <- ShearModelParameters[ModelID %in% MID, Ce]
        N <- ShearModelParameters[ModelID %in% MID, n]
        Fe <- sapply(Ce, function(x) {
            (x - eo[k])^2 / (1 + eo[k])
        })


        m1 <- ifelse(
            GID[k] == "Fines",
            approx(x = c(0, 20, 40, 60, 80, 100), y = c(0, 0.18, 0.30, 0.41, 0.48, 0.48), xout = IP[k])$y, 0
        )

        # pref <- 100 #[kPa]
        Gref <- sapply(seq(1, length(Fe)), function(n) {
            A[n] * Fe[n] * (OCR[k]^m1) * 100^(N[n])
        }) # [MPa]
        Go <- sapply(seq(1, length(Fe)), function(n) {
            Gref[n] * (pm[k] / 100)^(N[n])
        }) # [MPa]
        Gm[k] <- mean(Go) # [MPa]
        VSm[k] <- sqrt(9.81 * Gm[k] * 1000 / gsat[k]) # sqrt(kPa / (9.81 m/s2 kN/m3)) [m/s]
    }


    # Get properties ----
    VSa <- (cumsum(hs) / cumsum(hs / VSm)) |> round(digits = 1)
    VS30 <- .setVs30(hs = hs, VSm = VSm, VSo = Vref) |> round(digits = 1)
    SID <- Vs30toSID(VS30)
    # Fit Go,mo model ----

    # if(any(zm/Hs<=0) || any(Gm<=0) ){browser()}
    DATA <- (data.table(zm = zm, Gm = Gm))[Gm > 0][, .(X = log(zm / Hs), Y = log(Gm))]


    LM <- lm(formula = Y ~ X, data = DATA)
    LnGo <- LM$coefficients[1] |> unname()
    mo <- LM$coefficients[2] |>
        round(digits = 3) |>
        unname()
    Go <- exp(LnGo) |> round(digits = 2) # [MPa]


    # Gmax = rho* Vs^2
    # LM <- lm(formula=Z~X+Y,data=DATA)
    # VSo <- predict.lm(LM,newdata = data.table(X=0,Y=log(Go))) |> exp() |> round(digits = 1)
    VSo <- sqrt(9.81 * Go * 1000 / gsat[NL]) |> round(digits = 1)

    Ts <- fitModel.Ts(VSm = VSm, hs = hs, zm = zm) |> round(digits = 3)
    USCS.ID <- UID |>
        unique() |>
        sort() |>
        paste0(collapse = ".")
    # browser()
    Coarse <- round(sum(hs[UID %in% ValidCoarse]) / sum(hs) * 100, digits = 0)
    Gravels <- round(sum(hs[UID %in% ValidGravels]) / sum(hs) * 100, digits = 0)
    Fines <- round(sum(hs[UID %in% ValidFines]) / sum(hs) * 100, digits = 0)
    Sands <- round(sum(hs[UID %in% ValidSands]) / sum(hs) * 100, digits = 0)
    Silts <- round(sum(hs[UID %in% ValidSilts]) / sum(hs) * 100, digits = 0)
    Clays <- round(sum(hs[UID %in% ValidClays]) / sum(hs) * 100, digits = 0)
    Organic <- round(sum(hs[UID %in% ValidOrganic]) / sum(hs) * 100, digits = 0)

    MAP <- .mapLayers(UID = UID, hs = hs)

    Water <- round(100 * Water)

    # if(any(zm/Hs<=0) || any(Gm<=0) ){browser()}
    DATA <- (data.table(zm = zm, VSm = VSm))[VSm > 0][, .(X = log(zm / Hs), Z = log(VSm))]
    # DATA <- data.table(X=log(zm/Hs),Z=log(VSm))

    LM <- lm(formula = X ~ Z, data = DATA)
    Z500 <- predict(LM, newdata = data.table(Z = log(500))) |>
        exp() |>
        round(digits = 1)
    Z1000 <- predict(LM, newdata = data.table(Z = log(1000))) |>
        exp() |>
        round(digits = 1)
    # browser()
    SN <- paste0(
        str_pad(string = (round(10 * Ts, digits = 0)), width = 2, side = "left", pad = 0),
        str_pad(string = Hs, width = 3, side = "left", pad = 0), SID
    )

    # Output -----

    SiteProperties <- data.table(Hs, Hw, NL, Z500, Z1000, SID, Go, mo, Ts, VSo, VS30, UID = USCS.ID, Gravels, Sands, Fines, Clays, Silts, Organic, Water, MAP, Go_Units = "MPa", Vs_Units = "m/s", SN, POP, POP_Units = "kPa")

    if (getSiteLayers) {
        SiteLayers <- data.table(
            USCS = UID[1:NL],
            GroupID = GID[1:NL],
            zm = zm[1:NL],
            hs = hs[1:NL],
            emin = emin[1:NL] |> round(digits = 2),
            eo = eo[1:NL] |> round(digits = 2),
            emax = emax[1:NL] |> round(digits = 2),
            gsmin = gsmin[1:NL] |> round(digits = 1),
            gs = gs[1:NL] |> round(digits = 1),
            gsat = gsat[1:NL] |> round(digits = 1),
            gsmax = gsmax[1:NL] |> round(digits = 1),
            gw = gw[1:NL],
            Dr = Dr[1:NL] |> round(digits = 2),
            IP = IP[1:NL] |> round(digits = 2),
            LL = LL[1:NL] |> round(digits = 2),
            pm = pm[1:NL] |> round(digits = 2),
            OCR = OCR[1:NL] |> round(digits = 1),
            Gm = (Gm[1:NL]) |> round(digits = 2), # MPa
            VSm = VSm[1:NL] |> round(digits = 1),
            VSa = VSa[1:NL] |> round(digits = 1),
            SID = SID
        )

        # browser()
        return(list(SiteLayers = SiteLayers, SiteProperties = SiteProperties))
    } else {
        return(SiteProperties)
    }
}





#
# .tagSites <- function(.SD){
#   sapply(seq(1,nrow(.SD)), function(n){digest::digest(object = .SD[n],algo="crc32")})
# }
#


.setVs30 <- function(hs, VSm, VSo) {
    ZTop <- c(0, head(cumsum(hs), -1)) # depth to top of each layer
    ZBot <- ZTop + hs # depth to bottom of each layer

    h30 <- pmax(pmin(ZBot, 30) - ZTop, 0) # thickness that lies in 0–30 m
    t30 <- h30 / VSm # travel time of logged layers

    if (sum(hs) < 30) { # pad with foundation if needed
        t30 <- c(t30, (30 - sum(hs)) / VSo)
    }

    round(30 / sum(t30), 1)
}

#' Site Class Data
#'
#' This dataset contains Site Class data from ASCE 7-22
#'
#' @format A data frame with 9 rows and 4 variables:
#' \describe{
#'   \item{SC}{Character: Site Class}
#'   \item{Description}{Character: Category description}
#'   \item{Vs30 (m/s)}{Character: Vs30 in m/s}
#'   \item{Vs30 (ft/s)}{Character: Vs30 in ft/s}
#' }
#' @usage data(SiteClass)
#' @docType data
#' @keywords datasets
#' @name SiteClass
"SiteClass"

#' Site-amplified AUXtral acceleration (Seyhan & Stewart 2014)
#'
#' Computes site-amplified AUXtral acceleration using two modes:
#' (A) If the input UHS contains only p == "mean", rock Sa is treated as
#'     deterministic and the dispersion in SaF comes solely from F_ST17
#'     (sdLnF), sampled on the natural-log scale.
#' (B) If the input UHS contains numeric quantiles (e.g., "0.10", "0.50", ...),
#'     ln Sa(Tn) is sampled from those quantiles using an empirical inverse
#'     quantile function (buildQSpline). Dependence between PGA and Sa(Tn)
#'     is represented by a Gaussian "star" copula (Baker & Jayaram, 2009)
#'     with correlation rhoBJ(Tn, T0 = 0, Rrup). For each Monte Carlo draw,
#'     F_ST17 is evaluated with the paired PGA draw, and ln SaF = ln Sa + ln F
#'     with ln F ~ Normal(muLnF, sdLnF). Independence between ln Sa and ln F
#'     is assumed.
#'
#' Requirements:
#' - The input table must have columns Tn, p, Sa.
#' - It must include exactly one row (Tn == 0, p == "mean") providing PGA on rock.
#'
#' Output columns: Tn, p, Sa, SaF, AF, where AF = SaF / Sa.
#' In mode (B), the same numeric p values found in the input are returned plus
#' "mean". In mode (A), the set of p values is taken from p_TARGET plus "mean".
#'
#' @param uhs       data.table with columns Tn, p, Sa.
#' @param vs30      target Vs30 (m/s).
#' @param vref      reference Vs30 (m/s), default 760.
#' @param ns        Monte Carlo size (default 1000).
#' @param models    character. Site-response model id. Currently the only
#'                  implemented model is \code{"ST17"}; reserved for future
#'                  site-response models.
#' @param Rrup      rupture distance (km) for rhoBJ() in the quantile case (default 100).
#' @param p_TARGET  probabilities to report when only 'mean' is provided,
#'                  default c(0.05, 0.10, 0.16, 0.50, 0.84, 0.90, 0.95).
#'
#' @return data.table with columns Tn, p, Sa, SaF, AF.
#' @examples
#' \dontrun{
#' uhs <- data.table::fread(
#'   system.file("extdata", "uhs.csv", package = "newmark"))
#' fitSaF(uhs, vs30 = 360, vref = 760, ns = 500)
#' }
#' @export
fitSaF <- function(uhs,
                   vs30,
                   vref = 760,
                   ns = 1000,
                   models = "ST17",
                   Rrup = 100,
                   p_TARGET = c(0.05, 0.10, 0.16, 0.50, 0.84, 0.90, 0.95)) {

  stopifnot(data.table::is.data.table(uhs))
  if (!all(c("Tn", "p", "Sa") %in% names(uhs))) {
    stop("fitSaF: 'uhs' must have columns Tn, p, Sa")
  }

  ## PGA (roca, media única)
  PGA <- uhs[Tn == 0 & p == "mean", Sa]
  if (length(PGA) != 1L) {
    stop("fitSaF: need exactly one (Tn == 0, p == 'mean') row for PGA")
  }

  ## Detección del tipo de entrada (ramificación única)
  PNumAll <- suppressWarnings(as.numeric(uhs$p))
  HaveQuantiles <- any(!is.na(PNumAll) & PNumAll > 0 & PNumAll < 1)

  Tn <- sort(unique(uhs$Tn))
  OUT <- data.table::data.table(Tn = numeric(0), p = character(0),
                                Sa = numeric(0), SaF = numeric(0))

  if (!HaveQuantiles) {
    # ---------------------- (A) sólo p == "mean" -----------------------
    for (t in Tn) {
      Sa <- uhs[Tn == t & p == "mean", Sa]
      if (length(Sa) != 1L) {
        next  # Manejar Tns sin datos
      }

      AUX  <- F_ST17(PGA = PGA, Tn = t, vs30 = vs30, vref = vref)
      muLnF <- AUX$muLnF
      sdLnF <- AUX$sdLnF

      X <- stats::rnorm(ns, mean = muLnF, sd = sdLnF)  # ln F
      Z <- log(Sa) + X                                 # ln SaF

      OUT <- data.table::rbindlist(list(
        OUT,
        data.table::data.table(
          Tn  = t,
          p   = sprintf("%.2f", p_TARGET),
          Sa  = Sa,
          SaF = as.numeric(stats::quantile(exp(Z), p_TARGET, type = 7))
        ),
        data.table::data.table(
          Tn  = t,
          p   = "mean",
          Sa  = Sa,
          SaF = mean(exp(Z))
        )
      ), use.names = TRUE)
    }

  } else {
    # ---------------------- (B) hay cuantiles numéricos -----------------
    ## Preparar el muestreo de PGA una única vez
    DP0 <- uhs[Tn == 0]
    p0  <- suppressWarnings(as.numeric(DP0$p))
    id0 <- which(!is.na(p0) & p0 > 0 & p0 < 1)
    DP0q <- DP0[id0, .(p = as.numeric(p), Sa)]
    DP0q <- DP0q[order(p)]
    LnQ0 <- buildQSpline(DP0q)  # u -> ln PGA

    for (t in Tn) {
      DT <- uhs[Tn == t]
      Sa <- DT[p == "mean", Sa]
      if (length(Sa) != 1L) {
        next  # Manejar Tns sin datos
      }

      ## inverso empírico en ln‑espacio para Sa(Tn)
      PNum <- suppressWarnings(as.numeric(DT$p))
      idx  <- which(!is.na(PNum) & PNum > 0 & PNum < 1)
      DTn <- DT[idx, .(p = as.numeric(p), Sa)]
      DTn <- DTn[order(p)]
      LnQn <- buildQSpline(DTn)  # u in (0,1) -> ln Sa(Tn)

      ## cópula gaussiana “estrella” PGA–Sa(Tn)
      rho <- rhoBJ(Tn = t, T0 = 0, Rrup = Rrup)
      Sig <- matrix(c(1, rho, rho, 1), nrow = 2L)
      Z2  <- mvtnorm::rmvnorm(ns, sigma = Sig)  # (Z0, Zn)
      U0  <- stats::pnorm(Z2[, 1L])
      U1  <- stats::pnorm(Z2[, 2L])

      ## mapeo a ln‑marginales y recentering
      LnPGASample <- LnQ0(U0)
      LnPGASample <- log(exp(LnPGASample) * (PGA / mean(exp(LnPGASample))))

      LnSaSample  <- LnQn(U1)
      LnSaSample  <- log(exp(LnSaSample) * (Sa / mean(exp(LnSaSample))))

      ## ST17 con PGA por muestreo
      AUX <- F_ST17(PGA = exp(LnPGASample), Tn = t, vs30 = vs30, vref = vref)
      LnFSample <- stats::rnorm(ns, mean = AUX$muLnF, sd = AUX$sdLnF)
      lnSaF <- LnSaSample + LnFSample

      ## mismos p que trae el hazard en este Tn + media
      OUT <- data.table::rbindlist(list(
        OUT,
        data.table::data.table(
          Tn  = t,
          p   = DT$p[idx],
          Sa  = DT$Sa[idx],
          SaF = as.numeric(stats::quantile(exp(lnSaF), PNum[idx], type = 7))
        ),
        data.table::data.table(
          Tn  = t,
          p   = "mean",
          Sa  = Sa,
          SaF = mean(exp(lnSaF))
        )
      ), use.names = TRUE)
    }
  }

  OUT[, AF := SaF / Sa]
  data.table::setorder(OUT, Tn, p)
  OUT[]
}

#' Highcharts-style cubic spline interpolation
#'
#' Drop-in replacement for \code{stats::approx}.
#' Uses the Catmull-Rom spline (with the same smoothing
#' parameter as Highcharts \code{type = "spline"}) instead of straight-line
#' segments.  The call signature and return value are identical to
#' \code{stats::approx}, so it can be substituted transparently.
#'
#' @param x Numeric vector of abscissa values.
#' @param y Numeric vector of ordinate values, same length as \code{x}.
#' @param xout Optional numeric vector of points where the spline is
#'   evaluated.  If \code{NULL}, a regular grid of length \code{n} is used.
#' @param n Integer. Number of points to generate when \code{xout} is
#'   \code{NULL}.  Default is 50.
#' @param rule Handling of points outside the range of \code{x}.  \code{1}
#'   returns \code{NA}; \code{2} repeats the nearest endpoint value.  Matches
#'   the semantics of \code{stats::approx}.
#' @param log Logical. If \code{TRUE}, interpolation is performed in
#'   log-log space.  Default is \code{FALSE}.
#' @param smoothing Numeric tension parameter used in the Catmull-Rom to
#'   Bezier conversion.  Default is 1.5 (Highcharts default).
#' @param ... Additional arguments ignored; present only for full signature
#'   compatibility with \code{stats::approx}.
#'
#' @return A list with components \code{x} and \code{y}, just like
#'   \code{stats::approx}.
#'
#' @seealso \code{\link[stats]{approx}}
#'
#' @examples
#' x <- c(0.05, 0.10, 0.20, 0.30, 0.50, 1.0, 2.0, 3.0)
#' y <- exp(-x)
#' Td <- seq(0.05, 3.0, length.out = 100)
#' approx.spline(x, y, xout = Td, log = TRUE)$y
#'
#' @export
approx.spline <- function(x, y = NULL,
                          xout = NULL,
                          n = 50,
                          rule = 1,
                          log = FALSE,
                          smoothing = 1.5,
                          ...) {
    if (is.null(y)) stop("y must be supplied")
    stopifnot(
        length(x) == length(y),
        length(x) >= 2,
        is.numeric(x), is.numeric(y)
    )

    ord <- order(x)
    x <- x[ord]
    y <- y[ord]

    ## -- handle log-log option ----------------------------------------------
    HasZero <- FALSE
    if (log) {
        if (x[1L] == 0) { # drop leading Tn = 0, keep to re-insert
            HasZero <- TRUE
            x0 <- x[1L]
            y0 <- y[1L]
            x <- x[-1L]
            y <- y[-1L]
        }
        if (any(x <= 0) || any(y <= 0)) {
            stop("log = TRUE requires strictly positive x and y, except for a single leading Tn = 0.")
        }
        tx <- log(x)
        ty <- log(y)
        toX <- exp
        toY <- exp
    } else {
        tx <- x
        ty <- y
        toX <- identity
        toY <- identity
    }

    npts <- length(tx)
    if (npts < 2L) {
        stop("Need at least two positive-period points to spline.")
    }

    ## -- Bezier control-point helper ----------------------------------------
    controls <- function(i) {
        p <- max(1, i - 1)
        nx <- min(npts, i + 1)

        dx1 <- tx[i] - tx[p]
        dy1 <- ty[i] - ty[p]
        dx2 <- tx[nx] - tx[i]
        dy2 <- ty[nx] - ty[i]
        denom <- abs(dx1) + abs(dx2)
        if (denom == 0) denom <- 1

        fL <- (abs(dx1) / denom) / smoothing
        fR <- (abs(dx2) / denom) / smoothing

        list(
            P0 = c(tx[i], ty[i]),
            P1 = c(tx[i] - fL * dx1, ty[i] - fL * dy1),
            P2 = c(tx[i] + fR * dx2, ty[i] + fR * dy2),
            P3 = c(tx[nx], ty[nx])
        )
    }

    ## -- choose evaluation abscissae ----------------------------------------
    if (!is.null(xout)) {
        xo <- if (log) ifelse(xout <= 0, NaN, log(xout)) else xout
    } else {
        xo <- seq(min(tx), max(tx), length.out = n)
    }

    ## -- locate interval for each xo ----------------------------------------
    idx <- findInterval(xo, tx, all.inside = FALSE)
    yout <- rep(NA_real_, length(xo))

    if (rule == 2L) {
        yout[idx == 0] <- ty[1L]
        yout[idx >= npts] <- ty[npts]
    }

    inside <- which(idx > 0 & idx < npts)
    if (length(inside)) {
        for (k in inside) {
            i <- idx[k]
            C <- controls(i)
            t <- (xo[k] - C$P0[1]) / (C$P3[1] - C$P0[1])
            yout[k] <- (1 - t)^3 * C$P0[2] +
                3 * (1 - t)^2 * t * C$P1[2] +
                3 * (1 - t) * t^2 * C$P2[2] +
                t^3 * C$P3[2]
        }
    }

    yout <- toY(yout)
    if (HasZero) { # splice (0, Sa0) back in
        xo <- c(x0, xo)
        yout <- c(y0, yout)
    }

    list(
        x = if (log) toX(xo) else xo,
        y = yout
    )
}

#' Interpolate Sa(p) at arbitrary period using log-log interpolation
#'
#' @param uhs data.table with columns Tn, p, Sa.
#' @param Tn numeric scalar — target period for interpolation.
#'
#' @return data.table with interpolated Tn, p, Sa
#' @examples
#' uhs <- data.table::data.table(
#'   Tn = c(0.2, 0.5, 1.0, 0.2, 0.5, 1.0),
#'   p  = c("0.16", "0.16", "0.16", "0.84", "0.84", "0.84"),
#'   Sa = c(0.8, 0.6, 0.3, 1.2, 0.9, 0.5)
#' )
#' interpolateSaTable(uhs, Tn = 0.3)
#' @export
interpolateSaTable <- function(uhs, Tn) {
  stopifnot(all(c("Tn", "p", "Sa") %in% names(uhs)))
  stopifnot(is.numeric(Tn), length(Tn) == 1L)

  if (anyDuplicated(uhs[, .(Tn, p)])) {
    warning("interpolateSaTable(): input has duplicate (Tn, p) rows; ",
            "values will be collapsed via ties = 'mean'. Disaggregate ",
            "by (siteID, ID, Vs30, TR) before calling to silence this.",
            call. = FALSE)
  }

  p <- sort(unique(uhs[p != "mean"]$p))
  if (length(p) == 0L) {
    stop("interpolateSaTable(): missing quantiles (p != 'mean')")
  }

  SaTable <- data.table::rbindlist(lapply(p, function(pj) {
    rows <- uhs[p == pj]
    if (nrow(rows) < 2L) {
      stop("interpolateSaTable(): need >=2 Tn values for p = ", pj)
    }

    y <- suppressWarnings(stats::approx(
      x = log(rows$Tn + 1e-8),
      y = log(rows$Sa),
      xout = log(Tn),
      rule = 2
    ))$y

    data.table::data.table(Tn = Tn, p = pj, Sa = exp(y))
  }))

  MeanRow <- uhs[p == "mean"]
  if (nrow(MeanRow) >= 2L) {
    YMean <- suppressWarnings(stats::approx(
      x = log(MeanRow$Tn + 1e-8),
      y = log(MeanRow$Sa),
      xout = log(Tn),
      rule = 2
    ))$y

    SaTable <- data.table::rbindlist(list(
      SaTable,
      data.table::data.table(Tn = Tn, p = "mean", Sa = exp(YMean))
    ))
  } else if (nrow(MeanRow) == 1L) {
    SaTable <- data.table::rbindlist(list(
      SaTable,
      data.table::data.table(Tn = Tn, p = "mean", Sa = MeanRow$Sa)
    ))
  }

  data.table::setorder(SaTable, Tn, p)
  SaTable[]
}

#' SiteTable Data
#'
#' This dataset contains simulations of different site conditions
#'
#' @format A data frame with 255614 rows and 38 variables:
#' \describe{
#'   \item{Hs}{Double: Height in m.}
#'   \item{Hw}{Double: Water Table in m.}
#'   \item{NL}{Integer: Number of Layers}
#'  \item{Z500}{Double: Depth to 500 m/s in m.}
#'  \item{Z1000}{Double: Depth to 1000 m/s in m.}
#'  \item{SID}{String: Site ID}
#'   \item{Go}{Double: Shear Modulus in kPa.}
#'   \item{mo}{Double: Shear Modulus Exponent.}
#'   \item{Ts}{Double: Fundamental Period in s.}
#'   \item{VSo}{Double: Shear Wave Velocity in m/s.}
#'   \item{VS30}{Double: Shear Wave Velocity at 30 m in m/s.}
#'   \item{UID}{String: USCS ID}
#'   \item{Gravels}{Double: Percentage of Gravels.}
#'   \item{Sands}{Double: Percentage of Sands.}
#'   \item{Fines}{Double: Percentage of Fines.}
#'   \item{Clays}{Double: Percentage of Clays.}
#'   \item{Silts}{Double: Percentage of Silts.}
#'   \item{Organic}{Double: Percentage of Organic.}
#'   \item{Water}{Double: Percentage of Water.}
#'   \item{GP}{Double: GP Fraction.}
#'   \item{GM}{Double: GM Fraction.}
#'   \item{GC}{Double: GC Fraction.}
#'   \item{GW}{Double: GW Fraction.}
#'   \item{SP}{Double: SP Fraction.}
#'   \item{SM}{Double: SM Fraction.}
#'   \item{SC}{Double: SC Fraction.}
#'   \item{SW}{Double: SW Fraction.}
#'   \item{CL}{Double: CL Fraction.}
#'   \item{ML}{Double: ML Fraction.}
#'   \item{OL}{Double: OL Fraction.}
#'   \item{PT}{Double: PT Fraction.}
#'   \item{OH}{Double: OH Fraction.}
#'   \item{MH}{Double: MH Fraction.}
#'   \item{CH}{Double: CH Fraction.}
#'   \item{POP}{Double: Overpressure in Kpa.}
#'   \item{POP_Units}{String: Overpressure Units.}
#'   \item{Go_Units}{String: Shear Modulus Units.}
#'   \item{Vs_Units}{String: Shear Wave Velocity Units.}
#' }
#' @usage data(SiteTable)
#' @docType data
#' @keywords datasets
#' @name SiteTable
"SiteTable"

#' ShearModelParameters
#'
#' This dataset contains hear Model Parameters from Ishihara (1996)
#'
#' @format A data frame with 20 rows and 10 variables:
#' \describe{
#'  \item{ModelID}{Character: Model ID}
#'  \item{GroupID}{Character: Group ID}
#'  \item{NameID}{Character: NameID}
#'  \item{AuthorID}{Character: AuthorID}
#'  \item{emin}{Double: Min Void Ratio}
#'  \item{emax}{Double: Max Void Ratio}
#'  \item{A}{Double: A}
#'  \item{Ce}{Double: Ce}
#'  \item{n}{Double: Ce}
#'  \item{UN}{Character: Go Units}
#' }
#' @usage data(ShearModelParameters)
#' @docType data
#' @keywords datasets
#' @name ShearModelParameters
"ShearModelParameters"

#' Internal: Normalize investigation time vectors
#' @keywords internal
#' @noRd
.normalizeIt <- function(it, n) {
  IT <- as.numeric(it)
  if (!(length(IT) %in% c(1L, n))) {
    stop("Investigation time must be scalar or match the data length.")
  }
  if (length(IT) == 1L) {
    IT <- rep(IT, n)
  }
  if (any(!is.finite(IT) | IT <= 0, na.rm = TRUE)) {
    stop("Investigation time must be finite and > 0.")
  }
  IT
}

#' Internal: Convert POE to annual exceedance probability
#' @keywords internal
#' @noRd
poeToAep <- function(poe, it) {
  POE <- as.numeric(poe)
  IT <- .normalizeIt(it, length(POE))
  -log1p(-POE) / IT
}

#' Internal: Convert POE to return period
#' @keywords internal
#' @noRd
poeToTr <- function(poe, it) {
  POE <- as.numeric(poe)
  IT <- .normalizeIt(it, length(POE))
  IT / -log1p(-POE)
}

#' Internal: Convert annual exceedance probability to POE
#' @keywords internal
#' @noRd
aepToPoe <- function(aep, it) {
  AEP <- as.numeric(aep)
  IT <- .normalizeIt(it, length(AEP))
  -expm1(-IT * AEP)
}

#' Internal: Convert annual exceedance probability to return period
#' @keywords internal
#' @noRd
aepToTr <- function(aep) {
  1 / as.numeric(aep)
}

#' Build a monotone quantile spline Q(u) for ln(Sa)
#' @importFrom stats splinefun pnorm
#' @importFrom mvtnorm rmvnorm
#' @param SaTable data.table with probability column \code{p} (0-1 or "mean")
#'                and spectral-acceleration column \code{Sa}.
#' @return A closure \code{Q(u)} returning \code{ln(Sa)} for any \code{u in[0,1]}.
#' @examples
#' tbl <- data.table::data.table(p = c(0.16, 0.50, 0.84), Sa = c(0.3, 0.5, 0.8))
#' Q <- buildQSpline(tbl)
#' Q(0.5)   # median ln(Sa)
#' @export
buildQSpline <- function(SaTable) {
  requiredCols <- c("p", "Sa")
  if (!all(requiredCols %in% names(SaTable))) {
    stop("buildQSpline(): SaTable must contain columns 'p' and 'Sa'.")
  }

  DT <- data.table::as.data.table(SaTable)

  # discard 'mean' row
  DT <- DT[p != "mean"]

  if (nrow(DT) < 3L) {
    stop("buildQSpline(): 3 or more quantiles are required.")
  }

  if (anyDuplicated(DT$p)) {
    warning("buildQSpline(): input has duplicate p rows; values will ",
            "be collapsed by stats::splinefun. Disaggregate by ",
            "(siteID, ID, Vs30, TR) before calling to silence this.",
            call. = FALSE)
  }

  extraCols <- setdiff(names(DT), c("p", "Sa"))
  if (length(extraCols)) {
    bad <- extraCols[vapply(
      DT[, extraCols, with = FALSE],
      function(x) data.table::uniqueN(x) > 1L,
      logical(1)
    )]
    if (length(bad)) {
      stop(
        "buildQSpline(): non-unique columns: ",
        paste(bad, collapse = ", ")
      )
    }
  }

  pVec <- sort(as.numeric(DT$p))
  lnSaVec <- log(DT[order(as.numeric(p)), Sa])

  lnSaMono <- cummax(lnSaVec) # avoid descents
  if (any(diff(lnSaMono) <= 0)) { # break ties exactly
    lnSaMono <- lnSaMono + seq_along(lnSaMono) * 1e-8
  }

  splineFun <- suppressWarnings(stats::splinefun(
    x = pVec,
    y = lnSaMono,
    method = "hyman"
  ))

  Q <- function(u) {
    u <- pmin(pmax(u, 0), 1) # clamp to [0,1]
    splineFun(u)
  }

  Q
}

#' Monte Carlo site properties from synthetic soil profiles
#'
#' Generates `NR` realisations of synthetic soil profiles for a given
#' total height `Hs` and USCS classification, then summarises the
#' resulting Ts, mo, Go, VSo and VS30 by user-requested quantile
#' levels (and/or the mean). Each realisation is built by
#' \code{geSiteTable()}.
#'
#' @param Hs Numeric vector. Total height(s) to hard ground (m).
#' @param USCS Character vector. USCS soil classification codes
#'   (e.g. \code{c("GC", "CL", "ML")}).
#' @param POP Numeric. Pre-consolidation pressure (kPa). Default 100.
#' @param Water Numeric. Water-table depth as fraction of \code{Hs}.
#'   Default 0 (no water).
#' @param NR Integer. Number of Monte Carlo realisations. Default 1.
#' @param h Numeric. Layer thickness (m) in the discretisation.
#'   Default 1.00.
#' @param levels Vector. Quantile levels to report; entries can be
#'   numeric in (0, 1) and/or the literal \code{"mean"}.
#'   Default \code{c(0.05, 0.5, "mean", 0.95)}.
#' @param Vref Numeric. Reference shear-wave velocity (m/s) for
#'   site-class assignment. Default 760.
#'
#' @return data.table with one row per (Hs, POP, Hw, level)
#'   combination and columns USCS, Go, mo, Ts, VSo, VS30, Z500,
#'   Z1000, level. Numeric properties are rounded for readability.
#' @export getSiteProperties
#'
#' @import data.table
#' @importFrom stats predict
#'
#' @examples
#' \donttest{
#' getSiteProperties(
#'   Hs   = 30,
#'   USCS = c("GC", "CL", "ML"),
#'   NR   = 50,
#'   levels = c(0.16, "mean", 0.84)
#' )
#' }
getSiteProperties <- function(Hs, USCS, POP = 100, Water = 0, NR = 1, h = 1.00,
                              levels = c(0.05, 0.5, "mean", 0.95),
                              Vref = 760) {
  . <- NULL

  # Build dataset
  SiteTable <- data.table()


  for (pop in POP) {
    for (hs in Hs) {
      for (w in Water) {
        for (i in 1:NR) {
          if (interactive()) cat(sprintf("\rBuilding sample %d/%d..", i, NR))
          DT <- geSiteTable(Hs = hs, h = h, Water = w, USCS = USCS, POP = pop)
          SiteTable <- rbindlist(list(DT, SiteTable), use.names = TRUE)
        }
      }
    }
  }
  if (interactive()) cat("Done!\n")


  DT <- data.table()
  if (any(levels == "mean")) {
    DT <- SiteTable[, .(
      USCS = paste0(USCS, collapse = " "),
      Go = mean(Go) |> round(1),
      mo = mean(mo) |> round(3),
      Ts = mean(Ts) |> round(2),
      VSo = mean(VSo) |> round(1),
      VS30 = mean(VS30) |> round(1),
      Z500 = mean(Z500) |> round(1),
      Z1000 = mean(Z1000),
      level = "mean"
    ), by = .(Hs, POP, Hw)] |>
      rbind(DT)
  }
  if (any(levels != "mean")) {
    probs <- levels[levels != "mean"] |> as.numeric()
    DT <- SiteTable[, .(
      USCS = paste0(USCS, collapse = " "),
      Go = quantile(Go, probs = probs) |> round(1),
      mo = quantile(mo, probs = probs) |> round(3),
      Ts = quantile(Ts, probs = probs) |> round(2),
      VSo = quantile(VSo, probs = probs) |> round(1),
      VS30 = quantile(VS30, probs = probs) |> round(1),
      Z500 = quantile(Z500, probs = probs) |> round(1),
      Z1000 = quantile(Z1000, probs = probs) |> round(1),
      level = probs
    ), by = .(Hs, POP, Hw)] |>
      rbind(DT)
  }
  DT
}

---
title: "Getting started with newmark"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting started with newmark}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
```

## Overview

`newmark` implements a four-workflow pipeline for probabilistic seismic
performance analysis of slopes and embankments:

1. **Dynamic Site Response** — compute the fundamental period Ts and
   site-amplified UHS from the soil profile (`getSiteProperties`,
   `getCylinderRoots`, `fitSaF`).
2. **Hazard import** — import PSHA output from OpenQuake (`buildGMDP`).
3. **Displacement curves** — Monte Carlo Newmark ensemble Dn(ky)
   (`getDnKy`, `fitDnCurve`).
4. **Seismic coefficient** — invert ensemble draws to kmax(d*)
   (`invertDnDraws`).

This vignette demonstrates Workflows 3 and 4 using the bundled example
dataset. For the full pipeline see the
`pipeline` vignette (`vignette("pipeline", package = "newmark")`).

## Example dataset

`uhs.csv` is a site-amplified uniform-hazard spectrum included with the
package: NBCC hazard model, Vs30 = 560 m/s, TR = 10 000 yr, with quantile
levels 0.05–0.95 and mean (Mode B input).

```{r load}
library(newmark)
library(data.table)

uhs <- fread(system.file("extdata", "uhs.csv", package = "newmark"))
uhs[p %in% c("0.16", "mean", "0.84") & Tn <= 0.2, .(Tn, p, Sa)]
```

## Parameters

```{r params}
# Ts: fundamental period of the sliding mass (s).
# In production, derived from getSiteProperties() + getCylinderRoots()
# using the soil USCS profile and slope geometry (Ishihara 1996,
# Gazetas & Dakoulas 1985). Here set to a representative value.
Ts <- 0.60

# Mw: moment magnitude from PSHA disaggregation.
Mw <- 6.8

# Ensemble weights (0 = model inactive).
weights <- c(AM88 = 1, JB07 = 0, BT07 = 1, SR08 = 1, BM17 = 0, BM19 = 1)

# Displacement targets (cm).
Da <- c(0.5, 2.5, 5.0, 25.0)
```

## Workflow 3 — Displacement curve Dn(ky)

```{r fitDnCurve}
ky     <- getDnKy(uhs, Ts = Ts)
result <- fitDnCurve(
  uhs     = uhs,
  ky      = ky,
  Ts      = Ts,
  Mw      = Mw,
  NS      = 200,
  weights = weights
)

result$curve[IDn == "ensemble" & p == "mean", .(ky, Dn)]
```

## Workflow 4 — Seismic coefficient kmax(d*)

```{r invertDnDraws}
kmax <- invertDnDraws(result$draws, Da = Da, weights = weights)
kmax[p %in% c("0.16", "mean", "0.84")]
```

`kmax` is in g. The normalised pseudostatic coefficient is
Kh = kmax / PGA_rock x 100 %.

## Next steps

- `vignette("dynamic-site-response", package = "newmark")` — soil
  profile to fundamental period and site amplification
  (`getSiteProperties`, `geSiteTable`, `getCylinderRoots`,
  `fitModel.Ts`, `fitSaF`).
- `vignette("ensemble-formulation", package = "newmark")` —
  mathematical derivation of the probabilistic propagation.
- `vignette("pipeline", package = "newmark")` — the four-workflow
  overview at function level.
- Function reference: `?fitSaF`, `?getDnKy`, `?fitDnCurve`,
  `?invertDnDraws`, `?getSiteProperties`, `?getCylinderRoots`,
  `?buildGMDP`.

---
title: "Dynamic site response: fundamental period and stiffness profile"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Dynamic site response: fundamental period and stiffness profile}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>", eval = FALSE)
```

## Why this matters

The flexible-block Newmark displacement models (BT07, BM17, BM19)
evaluate spectral acceleration at a period proportional to the
fundamental period of the sliding mass *T*_s. Estimating *T*_s
therefore precedes any flexible-block analysis. The base
stiffness parameters required for the estimation — the maximum
shear modulus *G*_o and the shear-wave velocity *v*^o_S at the
embankment base — are established here from the soil profile via
Ishihara's (1996) shear-modulus model and the Gazetas–Dakoulas
(1985) inhomogeneous truncated shear-beam formulation.

This methodology was originally distributed as the `dsra` package
and was merged into `newmark` 1.1.0 (see `NEWS.md`). The functions
`getSiteProperties()`, `geSiteTable()`, `getCylinderRoots()`,
`fitModel.Ts()`, `Vs30toSID()` and `SIDtoVs30()`, together with
the reference datasets `CylinderRoots`, `ShearModelParameters`,
`SiteClass` and `SiteTable`, all originate from `dsra`.

---

## Fundamental period of an inhomogeneous truncated shear beam

For an embankment whose shear stiffness increases with depth
according to a power-law profile

$$
G(z) \;\approx\; G_o \left(\frac{z}{H_\text{max}}\right)^{m_o},
$$

the *j*-th modal period is

$$
T_s^{(j)} \;\simeq\; \frac{4\,\pi\,H_\text{max}}{a^{(j)}\,(2 - m_o)\,v_S^{o}},
$$

where

- *H*_max is the maximum height of the embankment;
- *a*^(*j*) is the *j*-th eigenvalue of the characteristic equation
  for the laterally constrained embankment under harmonic base
  excitation, depending jointly on *m*_o and the truncation ratio
  *λ*_o;
- *v*^o_S = √(*G*_o / ρ) is the shear-wave velocity at the base,
  with *G*_o the maximum shear modulus at depth *z* = *H*_max and
  ρ the bulk mass density.

For practical evaluation, the fundamental mode (*j* = 1) dominates
and the expression reduces to a function of the base shear-wave
velocity, the maximum height, and the two dimensionless geometric
ratios *m*_o and *λ*_o.

The truncation ratio *λ*_o captures the geometric influence of the
berm width *B*_o on the dynamic response:

$$
\lambda_o \;=\; \frac{B_o}{B_\text{max}} \;=\; \frac{B_o}{2\,H_\text{max}\,s + B_o},
$$

where *s* is the reciprocal of the average slope gradient
(tan β = 1 / *s*). A larger *λ*_o lengthens the fundamental
period.

The homogeneity ratio *m*_o reflects the power-law variation of
the shear modulus with depth. A larger *m*_o implies a steeper
stiffness gradient from surface to base and lengthens the
fundamental period relative to the uniform-stiffness case
(*m*_o = 0).

The eigenvalue *a*^(1) is computed from a precomputed
table of characteristic roots derived from the Gazetas–Dakoulas
(1985) characteristic equation:

```r
library(newmark)

an <- getCylinderRoots(
  mo          = 0.45,                 # homogeneity ratio
  lo          = 0.20,                 # truncation ratio
  no          = 1,                    # mode number
  model       = "nlm",                # interpolation: "lm", "nlm", "dt", "rf"
  extrapolate = TRUE                  # warn (do not clamp) outside the table
)
```

The reference table is the `CylinderRoots` dataset; its support
in (*m*_o, *λ*_o) is *m*_o ∈ [0, 0.95] and *λ*_o ∈ [0, 0.495]
across modes 1..8.

When the analyst already has a discrete *V*_S(*z*) profile (for
instance from CPT-derived correlations or downhole logging),
`fitModel.Ts()` computes *T*_s directly via Rayleigh's method
without going through the modal eigenvalue:

```r
Ts <- fitModel.Ts(
  VSm = c(220, 280, 340),               # shear-wave velocities by layer (m/s)
  hs  = c(10, 10, 10),                  # layer thicknesses (m)
  zm  = c(5, 15, 25)                    # layer mid-depths (m)
)
```

---

## Ishihara stiffness model

The maximum (small-strain) shear modulus at depth *z* is estimated
as a function of void ratio *e*_o and octahedral effective stress
σ′_o(*z*) using Ishihara's (1996) form:

$$
G(z,\,e_o) \;\approx\; A\,\frac{(C_e - 1)^{2}}{1 + e_o}\,\left(\frac{\sigma'_o(z)}{p_\text{ref}}\right)^{n},
$$

where

- *A* is a dimensionless material-dependent amplitude constant;
- *C*_e is a void-ratio shape parameter controlling the
  sensitivity of *G* to void ratio;
- *p*_ref is a reference pressure (typically atmospheric,
  101.3 kPa);
- *n* is the stress exponent governing the power-law dependence
  of stiffness on confining pressure.

Material parameters (*A*, *C*_e, *n*) are tabulated in the
package as `ShearModelParameters`, indexed by USCS soil class.
Void-ratio ranges by USCS class are also packaged (used
internally by the synthetic-profile generator):

```r
data(ShearModelParameters)            # 20 rows, parameters by ModelID
data(SiteTable)                       # ~256k rows, sampled site profiles
data(CylinderRoots)                   # ~307k rows, eigenvalues a(mo, lo, n)
data(SiteClass)                       # 9 rows, ASCE 7-22 site-class table
```

The maximum shear modulus *G*_o corresponds to *z* = *H*_max,
where the octahedral stress reaches its largest value under the
self-weight of the fill. The base shear-wave velocity follows
directly:

$$
v_S^{o} \;=\; \sqrt{\frac{G_o}{\rho}}.
$$

---

## Synthetic profile generation

For preliminary studies where laboratory data are limited, the
package generates Monte Carlo realisations of synthetic soil
profiles consistent with a USCS classification. Each realisation
samples void ratio, unit weight and plasticity from
USCS-conditioned distributions, computes the stiffness profile
*G*(*z*) via the Ishihara expression, fits the power-law form to
extract (*G*_o, *m*_o, *v*^o_S), and computes *T*_s via the modal
expression above.

```r
SiteProps <- getSiteProperties(
  Hs     = 30,                         # total height to hard ground (m)
  USCS   = c("GC", "CL", "ML"),        # USCS codes (top to bottom)
  POP    = 100,                        # pre-consolidation pressure (kPa)
  Water  = 0,                          # water-table depth as fraction of Hs
  NR     = 100,                        # Monte Carlo realisations
  h      = 1.00,                       # layer thickness (m)
  levels = c(0.05, 0.5, "mean", 0.95),
  Vref   = 760
)
# returns a data.table with one row per (Hs, POP, Hw, level) combination
# columns: USCS, Go, mo, Ts, VSo, VS30, Z500, Z1000, level
```

`geSiteTable()` is the lower-level entry point that produces a
single realisation; `getSiteProperties()` wraps it in a Monte
Carlo loop and returns quantile summaries.

---

## Site-class utilities

The eight-class site identifier of ASCE 7-22 / NBCC is mapped
bidirectionally with the equivalent *V*_S30:

```r
Vs30toSID(c(120, 360, 760, 1500))
#> [1] "E"  "CD" "BC" "A"

SIDtoVs30(c("BC", "C", "CD", "D"))
#> [1] 760 540 370 255
```

The canonical class boundaries are:

| Class | Vs30 range (m/s)    | Representative Vs30 |
|-------|---------------------|--------------------:|
| A     | ≥ 1500              | 1500                |
| B     | 900 – 1500          | 1200                |
| BC    | 640 – 900           | 760                 |
| C     | 440 – 640           | 540                 |
| CD    | 300 – 440           | 370                 |
| D     | 210 – 300           | 255                 |
| DE    | 150 – 210           | 180                 |
| E     | < 150               | 150                 |

---

## Application

Most TSF and waste-rock dump assessments call
`getSiteProperties()` once per (geometry, material) pair to
obtain *T*_s and *m*_o, then pass *T*_s to the Newmark
displacement workflow (`fitDnCurve`). The function returns a
distribution because each input parameter (void ratio, unit
weight, *G* profile coefficients) is sampled from
USCS-conditioned ranges; the analyst chooses a quantile of *T*_s
for the design.

For site amplification of the rock UHS to the target *V*_S30,
see `?fitSaF` and the methodology in
`vignette("ensemble-formulation", package = "newmark")`.

---

## References

- Gazetas, G. & Dakoulas, P. (1985). *Soil Dyn. Earthq. Eng.* 4(4):166–182.
- Ishihara, K. (1996). *Soil Behaviour in Earthquake Geotechnics*. Oxford University Press.
- ASCE/SEI 7-22 (2022). *Minimum Design Loads and Associated Criteria for Buildings and Other Structures*. American Society of Civil Engineers.

---
title: "Newmark ensemble formulation"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Newmark ensemble formulation}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>", eval = FALSE)
```

## Overview

`newmark` answers one question: given a site-specific seismic hazard and a
slope with known dynamic properties, what horizontal seismic coefficient
*k*_max is required to limit permanent co-seismic displacement to a
specified target *d*\*?

The performance-based design problem is formulated as: for a given
target tolerable displacement *d*\* > 0, target return period *T*_R,
and target exceedance probability *p* ∈ (0, 1), choose the smallest
yield coefficient *k*_y such that the displacement induced at *T*_R
remains below *d*\* with probability at least 1 − *p*:

$$
k_\text{max}(p \mid d^*, T_R) \;=\; \inf\!\Bigl\{\,k_y \,:\, P\!\bigl[\,D_N(k_y; T_R) > d^*\,\bigr] \,\leq\, p\,\Bigr\}.
$$

The right-hand side has no closed form — neither the joint
distribution of intensity measures nor the convolution with the
empirical regression residual admits one. The framework evaluates
the inverse mapping numerically via Monte Carlo, propagating two
sources of uncertainty:

1. **Hazard uncertainty** — spectral accelerations are drawn from the
   quantile curves of the uniform-hazard spectrum via a Gaussian copula that
   preserves the **full** inter-period correlation matrix (Baker &
   Jayaram 2008); no star/hub simplification.
2. **Model uncertainty** — six empirical Newmark models are combined in a
   logic-tree ensemble with user-assigned epistemic weights *w*_i, ∑*w*_i = 1.

For the full mathematical derivation and calibration discussion, see:

> Verri Kozlowski, A. (2026). *Probabilistic estimation of Newmark
> displacements and seismic coefficients under hazard uncertainty.*
> Working paper.

---

## Three-stage structure

### Stage 1 — Site amplification (`fitSaF`)

Rock-level spectral ordinates are amplified to the target site Vs30 using
the Seyhan & Stewart (2014) nonlinear site-factor model (ST17).

- **Mode A** (mean-only input): rock Sa is deterministic; dispersion comes
  from ST17 sigma alone.
- **Mode B** (quantile input): Sa is sampled jointly across periods via
  Gaussian copula; dispersion reflects combined hazard and site-factor
  uncertainty.

See `?fitSaF` for the function-level interface.

### Stage 2 — Displacement curve (`fitDnCurve`)

For each Monte Carlo realisation, one ground-motion scenario is drawn
(a consistent tuple of PGA, Sa(1.3Ts), Sa(1.5Ts)) and passed through all
active displacement models simultaneously using a shared aleatory residual.
The result is a family of coherent displacement curves *D*_N(*k*_y).

### Stage 3 — Coefficient inversion (`invertDnDraws`)

Each curve is projected monotone and inverted in log-log space to find
the yield acceleration *k*_max such that *D*_N(*k*_max) = *d*\*. Mean and
quantiles of *k*_max over all realisations constitute the output.

---

## Displacement models

| ID | Authors | Type | Spectral reference |
|---|---|---|---|
| AM88 | Ambraseys & Menu (1988) | rigid block | PGA |
| JB07 | Jibson (2007) | rigid block | PGA, Arias intensity |
| SR08 | Saygili & Rathje (2008) | rigid block | PGA, Arias intensity |
| BT07 | Bray & Travasarou (2007) | flexible block | Sa(1.5 Ts) |
| BM17 | Bray, Macedo & Travasarou (2018) | flexible block, subduction | Sa(1.5 Ts) |
| BM19 | Bray & Macedo (2019, corr. 2023) | flexible block, crustal | Sa(1.3 Ts) |

Activate/deactivate models via the `weights` argument (0 = inactive).

For BM19, near-fault pulse motions are flagged when PGV > 115 cm s⁻¹
and the equation switches to the Bray-Macedo (2023) D100 / D50 form
controlled by the `NFC` argument (`"D100"` is the maximum-component
default, conservative for slopes within ±45° of fault-normal;
`"D50"` is the median-component case for other orientations). A
sub-regime split at PGV = 150 cm s⁻¹ inside the pulse equation
captures the saturation of seismic displacement at very high PGV.

---

## Key assumptions

- Shared aleatory residual across models per realisation: a single
  *z*^(*n*) ∼ 𝒩(0,1) is drawn and applied to all six models scaled
  by each model's own σ_lnD. This corresponds to ρ = 1
  cross-model residual correlation. Independent residuals per model
  (ρ = 0) underestimates the joint ensemble variability;
  intermediate cross-model correlations are an open refinement
  direction.
- Independent site-factor and Sa uncertainty within each realisation.
- Log-log linear inversion with boundary extrapolation. Per-model
  calibration ranges (`getKyLimits()`) are reported as a diagnostic
  but not enforced as a clamp inside `invertDnDraws()`.

## Operational defaults

- Monte Carlo sample size *N*_S: routine practice is
  *N*_S ∈ \[10³, 10⁴\]. Convergence diagnostics (band-width
  stability under *N*_S × 2) accompany each application.
- *k*_y grid points: *N*_k = 30, log-spaced over
  \[0.01, max(PGA, 0.80)\] g (default of `getDnKy()`).
- Reported quantiles: mean plus *p* ∈ {0.16, 0.50, 0.84}.
  Extreme-tail reporting (*p* ∈ {0.05, 0.95}) is requested
  explicitly via the `p` argument to `invertDnDraws()`.

---

## References

- Baker & Jayaram (2008). *Earthquake Spectra* 24(1):299–317.
- Bray & Travasarou (2007). *J. Geotech. Geoenviron. Eng.* 133(4):381–392.
- Bray, Macedo & Travasarou (2018). *J. Geotech. Geoenviron. Eng.* 144(3):04017124.
- Bray & Macedo (2019, corr. 2023). *J. Geotech. Geoenviron. Eng.* 145(12); *Soil Dyn. Earthq. Eng.* 168:107835.
- Jibson (2007). *Engineering Geology* 91(2–4):209–218.
- Saygili & Rathje (2008). *J. Geotech. Geoenviron. Eng.* 134(6):790–803.
- Seyhan & Stewart (2014). *Earthquake Spectra* 30(3):1241–1256.

---
title: "Full pipeline overview"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Full pipeline overview}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>", eval = FALSE)
```

## Scope

`newmark` is an R package: every workflow in this overview reads
its inputs from R objects in memory and returns R objects in
memory (`data.table`s plus, for `getCylinderRoots()`, a numeric
scalar). No function writes to disk. Project-scale orchestration
on a hazard grid — looping over scenarios, persisting intermediate
tables, aggregating across models or sites — is the consumer's
responsibility; one example consumer is the `oqt` command-line
tool, but `newmark` itself does not depend on it and works
identically when called from any R session.

## The four workflows

```
[Workflow 1]  Dynamic site response
              soil profile + USCS  ─►  getSiteProperties()  ─►  Ts, mo, Go, VSo
              rock UHS  +  target Vs30  ─►  fitSaF()  ─►  site-amplified UHS

[Workflow 2]  Hazard import
              OpenQuake classical/disagg  ─►  buildGMDP()  ─►  AEPTable, UHSTable, RMwTable

[Workflow 3]  Displacement curves                          ◄─── Ts from W1
              site UHS  ─►  getDnKy()  ─►  ky grid           ◄─── UHS from W1+W2
              site UHS + ky + Ts + Mw  ─►  fitDnCurve()  ─►  list(curve, draws)

[Workflow 4]  Seismic coefficient
              draws + Da  ─►  invertDnDraws()  ─►  kmax(d*)
```

Workflows 1 and 2 are independent. Workflow 3 consumes Ts from
Workflow 1 and the site-amplified UHS from Workflows 1+2.
Workflow 4 consumes the per-realisation draws from Workflow 3.

## Workflow 1 — Dynamic site response

Two sub-steps. The first builds the soil profile and the
fundamental period; the second amplifies the rock UHS to the
target Vs30.

```r
library(newmark)
library(data.table)

# 1.1 — synthetic soil profile and dynamic properties
SiteProps <- getSiteProperties(
  Hs     = 30,                          # total height to hard ground (m)
  USCS   = c("GC", "CL", "ML"),         # USCS codes, top to bottom
  h      = 1.00,                        # discretisation step (m)
  NR     = 100,                         # Monte Carlo realisations
  levels = c(0.16, "mean", 0.84)
)
# returns a data.table with Hs, level, Go, mo, Ts, VSo, VS30, ...

Ts  <- SiteProps[level == "mean", Ts]
mo  <- SiteProps[level == "mean", mo]
VSo <- SiteProps[level == "mean", VSo]

# 1.2 — site amplification to a target Vs30
saf <- fitSaF(
  uhs   = uhs_rock,                     # rock UHS data.table (Tn, p, Sa)
  vs30  = 350,                          # target site Vs30 (m/s)
  vref  = 760,                          # reference rock Vs30
  ns    = 1000,                         # Monte Carlo size
  Rrup  = 100                           # rupture distance (km)
)
# returns a data.table with Tn, p, Sa, SaF, AF
```

Two reference vignettes deepen this workflow:
`vignette("dynamic-site-response", package = "newmark")` covers
the methodology side (Ishihara, Gazetas–Dakoulas, ST17);
function-level docs are at `?getSiteProperties`,
`?getCylinderRoots`, `?fitModel.Ts`, `?Vs30toSID`, `?SIDtoVs30`,
`?fitSaF`.

## Workflow 2 — Hazard import

Convert OpenQuake classical-PSHA and disaggregation output into
the table shapes consumed downstream.

```r
gmdp <- buildGMDP(
  path    = "/path/to/openquake/output",
  IDo     = "GMM",
  engine  = "openquake",
  vref    = 760,
  TRo     = c(100, 200, 500, 1000, 2000, 2500, 5000, 10000)
)
# returns a list with $AEPTable, $UHSTable, $RMwTable
```

`buildGMDP()` accepts both `engine = "openquake"` (zipped XML
output) and `engine = "user"` (Excel files supplied by the
analyst). See `?buildGMDP` for the full signature.

## Workflow 3 — Displacement curves

Build the *k*_y grid and run the coherent Monte Carlo over the
six-model ensemble.

```r
ky     <- getDnKy(saf, Ts = Ts)         # default 30 log-spaced points
result <- fitDnCurve(
  uhs     = saf,                        # site-amplified UHS
  ky      = ky,
  Ts      = Ts,
  Mw      = 6.8,                        # representative magnitude
  NS      = 200,                        # Monte Carlo size
  Rrup    = 100,
  weights = c(AM88 = 1, JB07 = 0, BT07 = 1, SR08 = 1, BM17 = 0, BM19 = 1),
  NFC     = "D100"                      # near-fault component selector for BM19
)
# result$curve  data.table(ky, p, Dn, IDn, w)  with quantiles + ensemble per ky
# result$draws  data.table(ky, s, Dn, IDn)     all NS realisations per model
```

The single-scenario sampling in `fitDnCurve()` (one PGA, Sa(1.3
*T*_s), Sa(1.5*T*_s) tuple drawn per realisation, correlated by
the full Baker-Jayaram inter-period matrix) and the shared *z*^(*n*)
across models (ρ = 1) are explained in
`vignette("ensemble-formulation", package = "newmark")`.

## Workflow 4 — Seismic coefficient

Invert the per-realisation draws to the design *k*_max at the
chosen displacement targets.

```r
kmax <- invertDnDraws(
  result$draws,
  Da      = c(0.5, 2.5, 5.0, 25.0),     # target displacements (cm)
  weights = c(AM88 = 1, JB07 = 0, BT07 = 1, SR08 = 1, BM17 = 0, BM19 = 1),
  p       = c(0.16, 0.84)                # report quantiles alongside the mean
)
# returns a data.table(Da, p, kmax) — kmax in g
```

The normalised pseudostatic coefficient is
*K*_h = *k*_max / mean rock-level PGA, expressed as a percentage.

## See also

- `vignette("newmark-quickstart", package = "newmark")` — minimal
  Workflows 3 + 4 example with the bundled NBCC UHS.
- `vignette("dynamic-site-response", package = "newmark")` — the
  dsra-side methodology for Ts and the site-amplification model.
- `vignette("ensemble-formulation", package = "newmark")` —
  mathematical derivation of the probabilistic propagation,
  monotone projection, and inversion.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fitModel.Ts.R
\name{fitModel.Ts}
\alias{fitModel.Ts}
\title{Site's Fundamental Period}
\usage{
fitModel.Ts(VSm, hs, zm)
}
\arguments{
\item{VSm}{Vector. Shear Wave Velocity profile in m/s (vector)}

\item{hs}{Vector. Layer Thickness profile in m (vector)}

\item{zm}{Vector. Layer Coordinates profile in m (vector)}
}
\value{
Double. Site's fundamental period Ts in seconds
}
\description{
Site's Fundamental Period
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ShearModelParameters.R
\docType{data}
\name{ShearModelParameters}
\alias{ShearModelParameters}
\title{ShearModelParameters}
\format{
A data frame with 20 rows and 10 variables:
\describe{
\item{ModelID}{Character: Model ID}
\item{GroupID}{Character: Group ID}
\item{NameID}{Character: NameID}
\item{AuthorID}{Character: AuthorID}
\item{emin}{Double: Min Void Ratio}
\item{emax}{Double: Max Void Ratio}
\item{A}{Double: A}
\item{Ce}{Double: Ce}
\item{n}{Double: Ce}
\item{UN}{Character: Go Units}
}
}
\usage{
data(ShearModelParameters)
}
\description{
This dataset contains hear Model Parameters from Ishihara (1996)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getSiteProperties.R
\name{getSiteProperties}
\alias{getSiteProperties}
\title{Monte Carlo site properties from synthetic soil profiles}
\usage{
getSiteProperties(
  Hs,
  USCS,
  POP = 100,
  Water = 0,
  NR = 1,
  h = 1,
  levels = c(0.05, 0.5, "mean", 0.95),
  Vref = 760
)
}
\arguments{
\item{Hs}{Numeric vector. Total height(s) to hard ground (m).}

\item{USCS}{Character vector. USCS soil classification codes
(e.g. \code{c("GC", "CL", "ML")}).}

\item{POP}{Numeric. Pre-consolidation pressure (kPa). Default 100.}

\item{Water}{Numeric. Water-table depth as fraction of \code{Hs}.
Default 0 (no water).}

\item{NR}{Integer. Number of Monte Carlo realisations. Default 1.}

\item{h}{Numeric. Layer thickness (m) in the discretisation.
Default 1.00.}

\item{levels}{Vector. Quantile levels to report; entries can be
numeric in (0, 1) and/or the literal \code{"mean"}.
Default \code{c(0.05, 0.5, "mean", 0.95)}.}

\item{Vref}{Numeric. Reference shear-wave velocity (m/s) for
site-class assignment. Default 760.}
}
\value{
data.table with one row per (Hs, POP, Hw, level)
combination and columns USCS, Go, mo, Ts, VSo, VS30, Z500,
Z1000, level. Numeric properties are rounded for readability.
}
\description{
Generates \code{NR} realisations of synthetic soil profiles for a given
total height \code{Hs} and USCS classification, then summarises the
resulting Ts, mo, Go, VSo and VS30 by user-requested quantile
levels (and/or the mean). Each realisation is built by
\code{geSiteTable()}.
}
\examples{
\donttest{
getSiteProperties(
  Hs   = 30,
  USCS = c("GC", "CL", "ML"),
  NR   = 50,
  levels = c(0.16, "mean", 0.84)
)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rhoBJ.R
\name{rhoBJ}
\alias{rhoBJ}
\title{Baker and Jayaram (2008) inter-period correlation model}
\usage{
rhoBJ(Tn, T0 = 0.01, Rrup = 100)
}
\arguments{
\item{Tn}{Numeric scalar. Target spectral period in seconds.}

\item{T0}{Numeric scalar. Reference period in seconds. Default 0.01
(used as a numerical proxy for PGA). Values <= 0 are silently
converted to 0.01.}

\item{Rrup}{Numeric scalar. Not used; retained for backward compatibility.}
}
\value{
Numeric scalar correlation coefficient rho(T0, Tn).
}
\description{
Computes the Pearson correlation coefficient between the residuals of
ln(Sa) at two oscillator periods, using the canonical predictive equation
of Baker and Jayaram (2008), Earthquake Spectra 24(1), 299-317
(DOI 10.1193/1.2857544), Eqs. 5 and 6 verbatim.
}
\details{
The model has no dependence on rupture distance or magnitude.
\code{Rrup} is retained for signature compatibility and is silently ignored.
Periods of zero are treated as 0.01 s (standard PGA proxy in B&J 2008).

Form (Eqs. 5-6 of B&J 2008):
C1 = 1 - cos(pi/2 - 0.366 * ln(Tmax / max(Tmin, 0.109)))
C2 = (1 - 0.105 * (1 - 1/(1 + exp(100*Tmax - 5))) * (Tmax-Tmin)/(Tmax-0.0099))
if Tmax < 0.2; else 0
C3 = C2 if Tmax < 0.109; else C1
C4 = C1 + 0.5 * (sqrt(C3) - C3) * (1 + cos(pi * Tmin / 0.109))

rho = C2          if Tmax < 0.109
= C1          else if Tmin > 0.109
= min(C2, C4) else if Tmax < 0.2
= C4          otherwise

Validity: 0.01 <= T1, T2 <= 10 s.
}
\examples{
rhoBJ(Tn = 0.3, T0 = 0.01)   # rho(PGA, 0.3 s)
rhoBJ(Tn = 1.0, T0 = 0.5)

}
\references{
Baker, J. W. & Jayaram, N. (2008). Correlation of Spectral Acceleration
Values from NGA Ground Motion Models. Earthquake Spectra, 24(1), 299-317.
DOI: 10.1193/1.2857544.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/approx.spline.R
\name{approx.spline}
\alias{approx.spline}
\title{Highcharts-style cubic spline interpolation}
\usage{
approx.spline(
  x,
  y = NULL,
  xout = NULL,
  n = 50,
  rule = 1,
  log = FALSE,
  smoothing = 1.5,
  ...
)
}
\arguments{
\item{x}{Numeric vector of abscissa values.}

\item{y}{Numeric vector of ordinate values, same length as \code{x}.}

\item{xout}{Optional numeric vector of points where the spline is
evaluated.  If \code{NULL}, a regular grid of length \code{n} is used.}

\item{n}{Integer. Number of points to generate when \code{xout} is
\code{NULL}.  Default is 50.}

\item{rule}{Handling of points outside the range of \code{x}.  \code{1}
returns \code{NA}; \code{2} repeats the nearest endpoint value.  Matches
the semantics of \code{stats::approx}.}

\item{log}{Logical. If \code{TRUE}, interpolation is performed in
log-log space.  Default is \code{FALSE}.}

\item{smoothing}{Numeric tension parameter used in the Catmull-Rom to
Bezier conversion.  Default is 1.5 (Highcharts default).}

\item{...}{Additional arguments ignored; present only for full signature
compatibility with \code{stats::approx}.}
}
\value{
A list with components \code{x} and \code{y}, just like
\code{stats::approx}.
}
\description{
Drop-in replacement for \code{stats::approx}.
Uses the Catmull-Rom spline (with the same smoothing
parameter as Highcharts \code{type = "spline"}) instead of straight-line
segments.  The call signature and return value are identical to
\code{stats::approx}, so it can be substituted transparently.
}
\examples{
x <- c(0.05, 0.10, 0.20, 0.30, 0.50, 1.0, 2.0, 3.0)
y <- exp(-x)
Td <- seq(0.05, 3.0, length.out = 100)
approx.spline(x, y, xout = Td, log = TRUE)$y

}
\seealso{
\code{\link[stats]{approx}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/F_models.R
\name{SaF_ST17}
\alias{SaF_ST17}
\title{Site-amplified spectral acceleration using Seyhan & Stewart (2014)}
\usage{
SaF_ST17(Sa, pga, Tn, vs30, vref = 760)
}
\arguments{
\item{Sa}{numeric vector - rock spectral accelerations at period Tn (g)}

\item{pga}{numeric vector - peak ground acceleration on rock (g), same length as Sa}

\item{Tn}{numeric scalar - oscillator period (s)}

\item{vs30}{numeric scalar - time-averaged shear-wave velocity to 30 m (m/s)}

\item{vref}{numeric scalar - reference velocity (m/s), default 760}
}
\value{
A \link[data.table]{data.table} with columns:
\describe{
\item{\code{muLnSaF}}{natural-log mean of amplified Sa}
\item{\code{sdLnSaF}}{natural-log standard deviation of amplified Sa}
\item{\code{ID}}{character string "ST17"}}
}
\description{
Computes site-amplified spectral acceleration Sa_F using the Seyhan & Stewart (2014)
non-linear site factor model. This is the canonical implementation that calls
the underlying F_ST17() function.
}
\section{References}{

Seyhan, E. & Stewart, J.P. (2014) "Semi-empirical nonlinear site amplification
from NGA-West2 data and simulations." \emph{Earthquake Spectra} \strong{30}(3):1241--1256.
\doi{10.1193/063013EQS181M}
}

\examples{
library(data.table)

# Site amplification for soft soil site
Sa_rock <- c(0.1, 0.2, 0.3, 0.4)
PGA_rock <- c(0.1, 0.1, 0.1, 0.1)

result <- SaF_ST17(
  Sa = Sa_rock,
  pga = PGA_rock,
  Tn = 0.5,      # 0.5s period
  vs30 = 300,    # Soft soil
  vref = 760     # Rock reference
)

print(result)
}
\seealso{
\code{\link{F_ST17}} for the underlying site factor model

\code{\link{fitSaF}} for Monte-Carlo site amplification analysis
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fitDnCurve.R
\name{fitDnCurve}
\alias{fitDnCurve}
\title{Newmark displacement curve Dn(ky) for one scenario}
\usage{
fitDnCurve(uhs, ky, Ts, Mw = 6.5, NS = 30L, Rrup = 100, weights, NFC = "D100")
}
\arguments{
\item{uhs}{data.table with columns Tn, p, Sa (one scenario)}

\item{ky}{numeric vector, yield accelerations (g)}

\item{Ts}{numeric scalar, fundamental period (s)}

\item{Mw}{numeric scalar, moment magnitude}

\item{NS}{integer, Monte Carlo samples for curve quantiles and
per-realisation draws.}

\item{Rrup}{numeric scalar, rupture distance (km)}

\item{weights}{named numeric vector, ensemble weights by IDn}

\item{NFC}{character, near-fault component selector for model BM19:
\code{"D100"} (maximum component, default) or \code{"D50"}.
Ignored by all other submodels.}
}
\value{
list(curve, draws). curve = data.table(ky, p, Dn, IDn, w).
draws = data.table(ky, s, Dn, IDn) — always populated (possibly
empty when no model produces output). Dn in cm.
}
\description{
Evaluates ensemble and per-submodel Dn over a vector of ky values.
The scenario is sampled once and the same epsilon is reused across all ky
so that each realisation s represents a coherent physical curve Dn_s(ky).
}
\examples{
\dontrun{
uhs <- data.table::fread(
  system.file("extdata", "uhs.csv", package = "newmark"))
out <- fitDnCurve(uhs, ky = getDnKy(uhs, Ts = 0.3), Ts = 0.3, Mw = 7.5,
                  NS = 100, weights = c(AM88=1, BT07=1, SR08=1, BM17=1))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Dn_models.R
\name{Dn_JB07}
\alias{Dn_JB07}
\title{Jibson (2007) empirical model}
\usage{
Dn_JB07(PGA, ky, AI = NULL)
}
\arguments{
\item{PGA}{Peak ground acceleration (g).}

\item{ky}{Yield acceleration (g).}

\item{AI}{Arias Intensity (m/s); if NULL, back-calculated from PGA.}
}
\value{
data.table(muLnD, sdLnD, ID = "JB07")
}
\description{
Jibson (2007) empirical model
}
\examples{
Dn_JB07(PGA = 0.5, ky = 0.1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Dn_models.R
\name{Dn_AM88}
\alias{Dn_AM88}
\title{Ambraseys & Menu (1988) rigid sliding-block model}
\usage{
Dn_AM88(PGA, ky)
}
\arguments{
\item{PGA}{Peak ground acceleration (g).}

\item{ky}{Yield acceleration (g).}
}
\value{
data.table(muLnD, sdLnD, ID = "AM88")
}
\description{
Ambraseys & Menu (1988) rigid sliding-block model
}
\examples{
Dn_AM88(PGA = 0.5, ky = 0.1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checkUHS.R
\name{checkUHS}
\alias{checkUHS}
\title{Check monotonicity and duplicates in UHSTable}
\usage{
checkUHS(UHSTable, epsMonot = 1e-06)
}
\arguments{
\item{UHSTable}{data.table with columns
\code{TR, Vs30, Vref, ID, Tn, p, Sa}.}

\item{epsMonot}{numeric tolerance (default 1e-6).}
}
\value{
invisibly \code{TRUE}; stops on error.
}
\description{
Ensures that, for every combination
\code{TR, Vs30, Vref, ID, Tn}, there is exactly
one row per probability \code{p} and that
\code{Sa(p)} is non‑decreasing.
}
\examples{
\dontrun{
# Requires a fully populated UHSTable produced by buildGMDP() or an
# equivalent upstream pipeline (must include the ID column that
# disaggregates per-quantile rows). Not bundled with the package.
checkUHS(UHSTable)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/buildQSpline.R
\name{buildQSpline}
\alias{buildQSpline}
\title{Build a monotone quantile spline Q(u) for ln(Sa)}
\usage{
buildQSpline(SaTable)
}
\arguments{
\item{SaTable}{data.table with probability column \code{p} (0-1 or "mean")
and spectral-acceleration column \code{Sa}.}
}
\value{
A closure \code{Q(u)} returning \code{ln(Sa)} for any \code{u in[0,1]}.
}
\description{
Build a monotone quantile spline Q(u) for ln(Sa)
}
\examples{
tbl <- data.table::data.table(p = c(0.16, 0.50, 0.84), Sa = c(0.3, 0.5, 0.8))
Q <- buildQSpline(tbl)
Q(0.5)   # median ln(Sa)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/designUHS.R
\name{designUHS}
\alias{designUHS}
\title{MCER / Design elastic spectrum for a Vs30/Vref slice (ASCE 7-22)}
\usage{
designUHS(UHS, TL = 8, spectrum = c("MCER", "DESIGN"))
}
\arguments{
\item{UHS}{data.table slice from UHSTable (single TR, p == "mean").}

\item{TL}{numeric, long-period transition (seconds). Default 8.}

\item{spectrum}{character, "MCER" (default) or "DESIGN" (2/3 * MCER).}
}
\value{
data.table with columns \verb{Tn, SaF}.
}
\description{
The input \code{UHS} slice must contain at least \code{Tn} and \code{SaF}.
Typical usage passes a single TR and \code{p == "mean"}.
}
\examples{
uhs <- data.table::data.table(Tn = c(0, 0.2, 0.5, 1, 2), SaF = c(0.4, 1.0, 0.9, 0.6, 0.3))
designUHS(uhs)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getKyLimits.R
\name{getKyLimits}
\alias{getKyLimits}
\title{Get calibrated ky range for each displacement model}
\usage{
getKyLimits(PGA)
}
\arguments{
\item{PGA}{numeric scalar, peak ground acceleration (g). Required for
rigid model limits.}
}
\value{
data.table with columns IDn, ky.lo, ky.hi
}
\description{
Returns the ky bounds within which each model's regression is calibrated.
For rigid models (AM88, SR08) the bounds are relative to PGA.
For flexible models the bounds are absolute.
}
\examples{
getKyLimits(PGA = 0.5)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SiteClass.R
\docType{data}
\name{SiteClass}
\alias{SiteClass}
\title{Site Class Data}
\format{
A data frame with 9 rows and 4 variables:
\describe{
\item{SC}{Character: Site Class}
\item{Description}{Character: Category description}
\item{Vs30 (m/s)}{Character: Vs30 in m/s}
\item{Vs30 (ft/s)}{Character: Vs30 in ft/s}
}
}
\usage{
data(SiteClass)
}
\description{
This dataset contains Site Class data from ASCE 7-22
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Dn_models.R
\name{Dn_BT07}
\alias{Dn_BT07}
\title{Bray & Travasarou (2007) flexible sliding-block model}
\usage{
Dn_BT07(ky, Sa, Ts, Mw = 6.5)
}
\arguments{
\item{ky}{Yield acceleration (g).}

\item{Sa}{Spectral acceleration at 1.5 * Ts (g).}

\item{Ts}{Fundamental period (s).}

\item{Mw}{Moment magnitude.}
}
\value{
data.table(muLnD, sdLnD, ID = "BT07")
}
\description{
Bray & Travasarou (2007) flexible sliding-block model
}
\examples{
Dn_BT07(ky = 0.1, Sa = 0.8, Ts = 0.3, Mw = 7.0)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Dn_models.R
\name{Dn_BM17}
\alias{Dn_BM17}
\title{Bray & Macedo (2017) - subduction/interface events}
\usage{
Dn_BM17(ky, Sa, Ts, Mw = 7.5)
}
\arguments{
\item{ky}{Yield acceleration (g).}

\item{Sa}{Spectral acceleration at 1.5 * Ts (g).}

\item{Ts}{Fundamental period (s).}

\item{Mw}{Moment magnitude (default 7.5).}
}
\value{
data.table(muLnD, sdLnD, ID = "BM17")
}
\description{
Bray & Macedo (2017) - subduction/interface events
}
\examples{
Dn_BM17(ky = 0.1, Sa = 0.8, Ts = 0.3, Mw = 8.0)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CylinderRoots.R
\docType{data}
\name{CylinderRoots}
\alias{CylinderRoots}
\title{Cylinder Roots Data}
\format{
A data frame with 307124 rows and 4 variables:
\describe{
\item{n}{Integer: Root number (1-10)}
\item{m}{Double: Inhomogeneity ratio (0-0.95)}
\item{l}{Double: Aspect Ratio (0-0.4950)}
\item{an}{Double: Cylinder roots}
}
}
\usage{
data(CylinderRoots)
}
\description{
This dataset contains cylinder roots.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/interpolateSaTable.R
\name{interpolateSaTable}
\alias{interpolateSaTable}
\title{Interpolate Sa(p) at arbitrary period using log-log interpolation}
\usage{
interpolateSaTable(uhs, Tn)
}
\arguments{
\item{uhs}{data.table with columns Tn, p, Sa.}

\item{Tn}{numeric scalar — target period for interpolation.}
}
\value{
data.table with interpolated Tn, p, Sa
}
\description{
Interpolate Sa(p) at arbitrary period using log-log interpolation
}
\examples{
uhs <- data.table::data.table(
  Tn = c(0.2, 0.5, 1.0, 0.2, 0.5, 1.0),
  p  = c("0.16", "0.16", "0.16", "0.84", "0.84", "0.84"),
  Sa = c(0.8, 0.6, 0.3, 1.2, 0.9, 0.5)
)
interpolateSaTable(uhs, Tn = 0.3)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SiteTable.R
\docType{data}
\name{SiteTable}
\alias{SiteTable}
\title{SiteTable Data}
\format{
A data frame with 255614 rows and 38 variables:
\describe{
\item{Hs}{Double: Height in m.}
\item{Hw}{Double: Water Table in m.}
\item{NL}{Integer: Number of Layers}
\item{Z500}{Double: Depth to 500 m/s in m.}
\item{Z1000}{Double: Depth to 1000 m/s in m.}
\item{SID}{String: Site ID}
\item{Go}{Double: Shear Modulus in kPa.}
\item{mo}{Double: Shear Modulus Exponent.}
\item{Ts}{Double: Fundamental Period in s.}
\item{VSo}{Double: Shear Wave Velocity in m/s.}
\item{VS30}{Double: Shear Wave Velocity at 30 m in m/s.}
\item{UID}{String: USCS ID}
\item{Gravels}{Double: Percentage of Gravels.}
\item{Sands}{Double: Percentage of Sands.}
\item{Fines}{Double: Percentage of Fines.}
\item{Clays}{Double: Percentage of Clays.}
\item{Silts}{Double: Percentage of Silts.}
\item{Organic}{Double: Percentage of Organic.}
\item{Water}{Double: Percentage of Water.}
\item{GP}{Double: GP Fraction.}
\item{GM}{Double: GM Fraction.}
\item{GC}{Double: GC Fraction.}
\item{GW}{Double: GW Fraction.}
\item{SP}{Double: SP Fraction.}
\item{SM}{Double: SM Fraction.}
\item{SC}{Double: SC Fraction.}
\item{SW}{Double: SW Fraction.}
\item{CL}{Double: CL Fraction.}
\item{ML}{Double: ML Fraction.}
\item{OL}{Double: OL Fraction.}
\item{PT}{Double: PT Fraction.}
\item{OH}{Double: OH Fraction.}
\item{MH}{Double: MH Fraction.}
\item{CH}{Double: CH Fraction.}
\item{POP}{Double: Overpressure in Kpa.}
\item{POP_Units}{String: Overpressure Units.}
\item{Go_Units}{String: Shear Modulus Units.}
\item{Vs_Units}{String: Shear Wave Velocity Units.}
}
}
\usage{
data(SiteTable)
}
\description{
This dataset contains simulations of different site conditions
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/buildGMDP.R
\name{buildGMDP}
\alias{buildGMDP}
\title{Build Ground-Motion Design Parameters (GMDP)}
\usage{
buildGMDP(
  path,
  IDo = "GEM",
  engine = "openquake",
  vref = 760,
  TRo = c(100, 200, 500, 1000, 2000, 2500, 5000, 10000)
)
}
\arguments{
\item{path}{Character. Directory with hazard data files.}

\item{IDo}{Character. Identifier for the GMDP. Default \code{"gmdp"}.}

\item{engine}{Character. \code{"openquake"} (default) or \code{"user"}.}

\item{vref}{Numeric. Reference Vs30 (m/s). Default 760.}

\item{TRo}{Numeric vector. Target return periods (years).}
}
\value{
A named list with \code{AEPTable}, \code{UHSTable}, and \code{RMwTable}
(\code{NULL} if disaggregation is unavailable).
}
\description{
Assembles an annual-exceedance-probability (AEP) table, uniform-hazard-spectrum
(UHS) table, and--where available--a magnitude-distance disaggregation table,
using OpenQuake or user-supplied hazard files.
}
\examples{
\dontrun{
# Requires an 'OpenQuake' Engine output directory containing classical
# PSHA + disaggregation ZIPs; not bundled with the package.
out <- buildGMDP(path = "path/to/openquake/output", vref = 760)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fitDnModel.R
\name{fitDnModel}
\alias{fitDnModel}
\title{Newmark displacement for a single submodel and scenario}
\usage{
fitDnModel(uhs, ky, Ts, Mw = 6.5, NS = 30L, Rrup = 100, IDn, NFC = "D100")
}
\arguments{
\item{uhs}{data.table with columns Tn, p, Sa (one scenario)}

\item{ky}{numeric scalar, yield acceleration (g)}

\item{Ts}{numeric scalar, fundamental period (s)}

\item{Mw}{numeric scalar, moment magnitude}

\item{NS}{integer, Monte Carlo samples}

\item{Rrup}{numeric scalar, rupture distance (km)}

\item{IDn}{character, submodel identifier}

\item{NFC}{character, near-fault component selector for model BM19:
\code{"D100"} (maximum component, default) or \code{"D50"}.
Ignored by all other submodels.}
}
\value{
data.table with columns p, Dn, IDn. Dn in cm.
}
\description{
Newmark displacement for a single submodel and scenario
}
\examples{
\dontrun{
uhs <- data.table::fread(
  system.file("extdata", "uhs.csv", package = "newmark"))
fitDnModel(uhs, ky = 0.1, Ts = 0.3, Mw = 7.5, NS = 100, IDn = "AM88")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getCylinderRoots.R
\name{getCylinderRoots}
\alias{getCylinderRoots}
\title{Characteristic root of an inhomogeneous truncated shear beam}
\usage{
getCylinderRoots(mo, lo, no = 1, model = "nlm", extrapolate = TRUE, OSF = 0.1)
}
\arguments{
\item{mo}{Double. Inhomogeneity ratio of the shear-modulus profile.
\code{mo = 0} is a homogeneous column. Range 0 to 0.95.}

\item{lo}{Double. Truncation ratio (top-radius / bottom-radius);
\code{lo = 0} is a fully truncated cone. Range 0 to 0.495.}

\item{no}{Integer. Mode number, 1..8. Default 1 (fundamental).}

\item{model}{Character. Interpolation model: \code{"lm"} (linear
GLM), \code{"nlm"} (non-linear GLM with interactions, default),
\code{"dt"} (decision tree), or \code{"rf"} (random forest).}

\item{extrapolate}{Logical. If \code{TRUE} (default), allow
\code{mo}/\code{lo} outside the table support and warn; if
\code{FALSE}, clamp to the table maxima.}

\item{OSF}{Double. Outlier scale factor used by the random-forest
model to shrink the training set around the query point. Default
0.10.}
}
\value{
Double. Characteristic root \code{an} for the requested
mode and parameters.
}
\description{
Returns the eigenvalue \code{an} of mode \code{no} for a Gazetas &
Dakoulas (1985) inhomogeneous truncated shear beam parameterised by
inhomogeneity ratio \code{mo} and truncation ratio \code{lo}. The
value is interpolated from the precomputed \code{CylinderRoots} table
using the requested model (linear glm, non-linear glm, decision
tree, or random forest).
}
\examples{
getCylinderRoots(mo = 0.45, lo = 0.20, no = 1, model = "nlm")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geSiteTable.R
\name{geSiteTable}
\alias{geSiteTable}
\title{Single-realisation synthetic soil profile and site properties}
\usage{
geSiteTable(
  Hs,
  Water = 0,
  USCS,
  Group = NULL,
  h = 0.5,
  DrID = NULL,
  Vref = 760,
  UniformDistribution = TRUE,
  POP = 100,
  IgnoreModelIntervals = TRUE,
  getSiteLayers = FALSE
)
}
\arguments{
\item{Hs}{Numeric scalar. Total height to hard ground (m).}

\item{Water}{Numeric scalar. Water-table depth as fraction of
\code{Hs}: \code{zw = Hs - Water * Hs}.}

\item{USCS}{Character vector. Unified Soil Classification System
codes (e.g. \code{c("GC", "CL", "ML")}).}

\item{Group}{Character vector or \code{NULL}. Soil groups,
\code{c("Gravels","Fines","Sands")}; if \code{NULL}, derived from
\code{USCS}.}

\item{h}{Numeric scalar. Layer thickness in the discretisation
(m). Default 0.50.}

\item{DrID}{Character or \code{NULL}. Relative-density label,
\code{c("Very Dense","Dense","Compact","Loose","Very Loose")};
\code{NULL} samples relative density from the standard prior.}

\item{Vref}{Numeric. Reference shear-wave velocity (m/s) for the
site-class assignment. Default 760.}

\item{UniformDistribution}{Logical. If \code{TRUE} (default),
sample soil properties from uniform distributions; if
\code{FALSE}, use normal distributions.}

\item{POP}{Numeric. Pre-consolidation pressure (kPa); enters the
over-consolidation ratio \code{OCR = (pm + POP) / pm}.
Default 100.}

\item{IgnoreModelIntervals}{Logical. If \code{TRUE} (default),
ignore model-specific layer-thickness intervals and use \code{h}
as the discretisation step.}

\item{getSiteLayers}{Logical. If \code{TRUE}, return the per-layer
detail in addition to the per-site summary.}
}
\value{
data.table with one row per (Hs, POP, Hw, sample)
combination and columns Hs, Hw, NL, Z500, Z1000, SID, Go, mo,
Ts, VSo, VS30, plus USCS-fraction columns.
}
\description{
Builds one Monte Carlo realisation of a layered soil column from a
total height \code{Hs} and a USCS classification, samples random
geotechnical properties (void ratio, unit weight, plasticity,
shear modulus parameters), computes the fundamental period
\code{Ts} via Rayleigh's method (\code{fitModel.Ts}) and the
Vs30-equivalent shear-wave velocity. \code{getSiteProperties}
calls this function repeatedly to assemble the Monte Carlo
ensemble; call directly when one realisation suffices.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getDnKy.R
\name{getDnKy}
\alias{getDnKy}
\title{Build a log-spaced ky grid for one scenario}
\usage{
getDnKy(uhs, Ts, kyN = 30L, pRef = "mean")
}
\arguments{
\item{uhs}{data.table with columns Tn, p, Sa (one scenario)}

\item{Ts}{numeric scalar, fundamental period (s)}

\item{kyN}{integer, number of grid points (default 30)}

\item{pRef}{character, probability level for intensity anchor (default "mean")}
}
\value{
numeric vector of ky values (log-spaced)
}
\description{
Derives the effective intensities PGA, Sa(1.3Ts), Sa(1.5Ts) from the
scenario UHS and constructs a log-spaced ky grid covering the calibrated
range of the displacement models.
}
\details{
Default bounds come from the calibration ranges of the models in the
ensemble:
\itemize{
\item lower: 0.01g (BM17 lowest calibration point)
\item upper: max(PGA, 0.80g). The physical limit is ky < PGA (no slip below
the yield acceleration); the 0.80g floor extends the grid for
low-PGA scenarios up to BM17's calibration ceiling.
}
}
\examples{
uhs <- data.table::data.table(Tn = 0, p = "mean", Sa = 0.5)
getDnKy(uhs, Ts = 0.3)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Dn_models.R
\name{Dn_SR08}
\alias{Dn_SR08}
\title{Saygili & Rathje (2008) model}
\usage{
Dn_SR08(PGA, ky, AI = NULL)
}
\arguments{
\item{PGA}{Peak ground acceleration (g).}

\item{ky}{Yield acceleration (g).}

\item{AI}{Arias Intensity (m/s); if NULL, back-calculated from PGA.}
}
\value{
data.table(muLnD, sdLnD, ID = "SR08")
}
\description{
Saygili & Rathje (2008) model
}
\examples{
Dn_SR08(PGA = 0.5, ky = 0.1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Dn_models.R
\name{Dn_BM19}
\alias{Dn_BM19}
\title{Bray & Macedo (2019, corrected 2023) - shallow-crustal events}
\usage{
Dn_BM19(ky, Ts, Sa, PGA, PGV, Mw = 6.5, NFC = "D100")
}
\arguments{
\item{ky}{Yield acceleration (g).}

\item{Ts}{Fundamental period (s).}

\item{Sa}{Spectral acceleration at 1.3 * Ts (g).}

\item{PGA}{Peak ground acceleration (g).}

\item{PGV}{Peak ground velocity (cm/s).}

\item{Mw}{Moment magnitude (default 6.5).}

\item{NFC}{Near-fault component: "D100" (default; max-component, slope
within 45 deg of fault-normal) or "D50" (median-component, other
orientations). Used only when PGV > 115 cm/s.}
}
\value{
data.table(muLnD, sdLnD, ID = "BM19")
}
\description{
Multi-equation structure per BM23 (Soil Dynamics and Earthquake Engineering
168, 107835, DOI 10.1016/j.soildyn.2023.107835):
Eq (10): ordinary motions, PGV <= 115 cm/s
Eq (6):  D100 max-component, PGV > 115 cm/s
Eq (7):  D50  median-component, PGV > 115 cm/s
with a sub-regime split at PGV = 150 cm/s within Eqs (6) and (7).
}
\examples{
Dn_BM19(ky = 0.1, Ts = 0.3, Sa = 0.8, PGA = 0.5, PGV = 80, Mw = 7.0)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sampleSaCorr.R
\name{sampleSaCorr}
\alias{sampleSaCorr}
\title{Draw a correlated sample of spectral accelerations}
\usage{
sampleSaCorr(uhs, Tn, rho, NS)
}
\arguments{
\item{uhs}{data.table with columns Tn, p, Sa (must include p == "mean").}

\item{Tn}{numeric vector; the first element is the reference period \code{Tn[1]},
e.g., 0.01 or 0.00. It does not have to be sorted.}

\item{rho}{numeric vector of length \code{length(Tn) - 1}; correlations between the
reference period \code{Tn[1]} and each remaining period.}

\item{NS}{integer, Monte Carlo sample size.}
}
\value{
matrix of dimension NS x length(Tn) with Sa draws (g).
}
\description{
Draw a correlated sample of spectral accelerations
}
\examples{
\dontrun{
uhs <- data.table::data.table(
  Tn = c(0, 0, 0, 1, 1, 1),
  p  = c("0.16", "mean", "0.84", "0.16", "mean", "0.84"),
  Sa = c(0.3, 0.5, 0.7, 0.1, 0.2, 0.3)
)
rho <- rhoBJ(Tn = 1.0, T0 = 0.01)
mat <- sampleSaCorr(uhs, Tn = c(0, 1), rho = rho, NS = 100)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SIDtoVs30.R
\name{SIDtoVs30}
\alias{SIDtoVs30}
\title{Convert Site Class to Vs30 in m/s}
\usage{
SIDtoVs30(SID = NULL)
}
\arguments{
\item{SID}{Site Class}
}
\value{
Vs30 in m/s
}
\description{
Convert Site Class to Vs30 in m/s
}
\examples{

SIDtoVs30("A")

SIDtoVs30("AB")

SIDtoVs30(c("BC","C","CD","D"))


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/F_models.R
\name{F_ST17}
\alias{F_ST17}
\title{Non-linear site amplification factor \strong{F\if{html}{\out{<sub>}}ST17\if{html}{\out{</sub>}}}}
\usage{
F_ST17(PGA, Tn, vs30, vref = 760, Vl = 200, Vu = 2000)
}
\arguments{
\item{PGA}{numeric \strong{vector} – peak ground acceleration (in g).}

\item{Tn}{numeric \strong{scalar} – oscillator period (s).}

\item{vs30}{numeric \strong{scalar} – time-averaged shear-wave velocity to 30 m (m/s).}

\item{vref}{numeric \strong{scalar} – reference velocity (760 m/s or 3000 m/s), default 760.}

\item{Vl}{numeric scalar – lower Vs30 bound for NL sigma interpolation (default 200).}

\item{Vu}{numeric scalar – upper Vs30 bound for NL sigma interpolation (default 2000).}
}
\value{
A \link[data.table]{data.table} with columns:
\describe{
\item{\code{muLnF}}{natural-log mean of \eqn{F}}
\item{\code{sdLnF}}{natural-log standard deviation of \eqn{F}
(set to \code{0} when \code{vs30 == vref})}
\item{\code{ID}}{character string \code{"ST17"}}}
}
\description{
Implements the period-dependent, Vs30-dependent model of Seyhan & Stewart (2014)
for the natural-log mean and standard deviation of the amplification factor
\eqn{F(T_n,PGA,V_{s30})}.
}
\details{
\emph{If \code{vs30 == vref}, the site factor equals 1 deterministically, so
the function returns \code{muLnF = 0} and \code{sdLnF = 0} (zero
dispersion; \code{rnorm(n, 0, 0)} returns zeros and \code{ln F = 0}).}
}
\section{References}{

Stewart, J.P. & Seyhan, E. (2017) “Semi-empirical nonlinear site
amplification model for global application.” \emph{Earthquake Spectra} \strong{33}(1).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Vs30toSID.R
\name{Vs30toSID}
\alias{Vs30toSID}
\title{Convert Vs30 (m/s) to Site Class Identifier}
\usage{
Vs30toSID(Vs30)
}
\arguments{
\item{Vs30}{Numeric vector. One or more Vs30 values in m/s.}
}
\value{
A character vector of the same length as \code{Vs30},
with site class designations.
}
\description{
Given one or more Vs30 values (in m/s), returns a character vector of site
class designations (\code{"A", "B", "BC", "C", "CD", "D", "DE", "E"}). The
classification thresholds are:
\itemize{
\item A: Vs30 >= 1500
\item B: 900 <= Vs30 < 1500
\item BC: 640 <= Vs30 < 900
\item C: 440 <= Vs30 < 640
\item CD: 300 <= Vs30 < 440
\item D: 210 <= Vs30 < 300
\item DE: 150 <= Vs30 < 210
\item E: 0 <= Vs30 < 150
}
}
\examples{
Vs30toSID(1500)  # returns "A"
Vs30toSID(c(120, 790, 3000, 455))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/invertDnDraws.R
\name{invertDnDraws}
\alias{invertDnDraws}
\title{Invert Dn draws to kmax(Da)}
\usage{
invertDnDraws(draws, Da, weights, p = c(0.05, 0.1, 0.16, 0.84, 0.9, 0.95))
}
\arguments{
\item{draws}{data.table(ky, s, Dn, IDn) — Dn in cm}

\item{Da}{numeric vector — displacement targets in cm}

\item{weights}{named numeric vector, ensemble weights by IDn}

\item{p}{numeric quantiles in (0,1) to report alongside the mean.
Default c(0.05, 0.10, 0.16, 0.84, 0.90, 0.95). The output
always includes the ensemble mean as the first p row.}
}
\value{
data.table(Da, p, kmax). Column \code{p} is character: "mean" or the
numeric quantile formatted as "0.84", "0.90", etc.
}
\description{
For each realisation s, inverts each model's Dn(ky) curve to kmax(Da),
builds weighted ensemble mean, then averages over realisations.
}
\details{
Log-log linear extrapolation is used when Da falls outside the support of
the computed Dn(ky) curve. No calibration-range clamping is applied — the
natural support of the curve bounds the inversion.

All inputs must be in the same displacement units (cm).
}
\examples{
\dontrun{
uhs <- data.table::fread(
  system.file("extdata", "uhs.csv", package = "newmark"))
out <- fitDnCurve(uhs, ky = getDnKy(uhs, Ts = 0.3), Ts = 0.3, Mw = 7.5,
                  NS = 100,
                  weights = c(AM88=1, BT07=1, SR08=1, BM17=1))
invertDnDraws(out$draws, Da = c(2.5, 5, 10),
              weights = c(AM88=1, BT07=1, SR08=1, BM17=1))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fitSaF.R
\name{fitSaF}
\alias{fitSaF}
\title{Site-amplified AUXtral acceleration (Seyhan & Stewart 2014)}
\usage{
fitSaF(
  uhs,
  vs30,
  vref = 760,
  ns = 1000,
  models = "ST17",
  Rrup = 100,
  p_TARGET = c(0.05, 0.1, 0.16, 0.5, 0.84, 0.9, 0.95)
)
}
\arguments{
\item{uhs}{data.table with columns Tn, p, Sa.}

\item{vs30}{target Vs30 (m/s).}

\item{vref}{reference Vs30 (m/s), default 760.}

\item{ns}{Monte Carlo size (default 1000).}

\item{models}{character. Site-response model id. Currently the only
implemented model is \code{"ST17"}; reserved for future
site-response models.}

\item{Rrup}{rupture distance (km) for rhoBJ() in the quantile case (default 100).}

\item{p_TARGET}{probabilities to report when only 'mean' is provided,
default c(0.05, 0.10, 0.16, 0.50, 0.84, 0.90, 0.95).}
}
\value{
data.table with columns Tn, p, Sa, SaF, AF.
}
\description{
Computes site-amplified AUXtral acceleration using two modes:
(A) If the input UHS contains only p == "mean", rock Sa is treated as
deterministic and the dispersion in SaF comes solely from F_ST17
(sdLnF), sampled on the natural-log scale.
(B) If the input UHS contains numeric quantiles (e.g., "0.10", "0.50", ...),
ln Sa(Tn) is sampled from those quantiles using an empirical inverse
quantile function (buildQSpline). Dependence between PGA and Sa(Tn)
is represented by a Gaussian "star" copula (Baker & Jayaram, 2009)
with correlation rhoBJ(Tn, T0 = 0, Rrup). For each Monte Carlo draw,
F_ST17 is evaluated with the paired PGA draw, and ln SaF = ln Sa + ln F
with ln F ~ Normal(muLnF, sdLnF). Independence between ln Sa and ln F
is assumed.
}
\details{
Requirements:
\itemize{
\item The input table must have columns Tn, p, Sa.
\item It must include exactly one row (Tn == 0, p == "mean") providing PGA on rock.
}

Output columns: Tn, p, Sa, SaF, AF, where AF = SaF / Sa.
In mode (B), the same numeric p values found in the input are returned plus
"mean". In mode (A), the set of p values is taken from p_TARGET plus "mean".
}
\examples{
\dontrun{
uhs <- data.table::fread(
  system.file("extdata", "uhs.csv", package = "newmark"))
fitSaF(uhs, vs30 = 360, vref = 760, ns = 500)
}
}
